import ctypes
import sys
from sdl2 import *


class PixelFormat:
    def __init__(self, pixel_format: int, app: any) -> None:
        self.pixel_format = pixel_format
        self.pixel_format_name = app.bts(SDL_GetPixelFormatName(self.pixel_format))
        self.pixel_type = SDL_PIXELTYPE(pixel_format)
        self.pixel_layout = SDL_PIXELLAYOUT(pixel_format)
        self.bits_per_pixel = SDL_BITSPERPIXEL(pixel_format)
        self.bytes_per_pixel = SDL_BYTESPERPIXEL(pixel_format)
        self.is_pixel_format_indexed = bool(SDL_ISPIXELFORMAT_INDEXED(pixel_format))
        self.is_pixel_format_alpha = bool(SDL_ISPIXELFORMAT_ALPHA(pixel_format))
        self.is_pixel_format_fourcc = bool(SDL_ISPIXELFORMAT_FOURCC(pixel_format))


class DisplayMode:
    def __init__(self, mode: SDL_DisplayMode = SDL_DisplayMode(), app: any = None) -> None:
        self.format = PixelFormat(mode.format, app)
        self.w = mode.w
        self.h = mode.h
        self.refresh_rate = mode.refresh_rate

    def as_dm(self) -> SDL_DisplayMode:
        return SDL_DisplayMode(self.format.pixel_format, self.w, self.h, self.refresh_rate)


class Display:
    def __init__(self, display_id: int, app: any) -> None:
        self.display_id = display_id
        self.name = app.bts(SDL_GetDisplayName(display_id))
        self.modes = []
        mode = SDL_DisplayMode()
        for i in range(SDL_GetNumDisplayModes(display_id)):
            SDL_GetDisplayMode(display_id, i, mode)
            self.modes.append(DisplayMode(mode, app))
        SDL_GetCurrentDisplayMode(self.display_id, mode)
        self.current_mode = DisplayMode(mode, app)
        SDL_GetDesktopDisplayMode(self.display_id, mode)
        self.desktop_mode = DisplayMode(mode, app)
        bounds = SDL_Rect()
        SDL_GetDisplayBounds(display_id, bounds)
        self.bounds = (bounds.x, bounds.y, bounds.w, bounds.h)
        dpi_ptr = [ctypes.c_float(0.0), ctypes.c_float(0.0), ctypes.c_float(0.0)]
        SDL_GetDisplayDPI(display_id, *dpi_ptr)
        self.d_dpi = dpi_ptr[0].value
        self.h_dpi = dpi_ptr[1].value
        self.v_dpi = dpi_ptr[2].value
        orientation_id = SDL_GetDisplayOrientation(display_id)
        if orientation_id == SDL_ORIENTATION_LANDSCAPE:
            self.orientation = 'landscape'
        elif orientation_id == SDL_ORIENTATION_LANDSCAPE_FLIPPED:
            self.orientation = 'landscape_flipped'
        elif orientation_id == SDL_ORIENTATION_PORTRAIT:
            self.orientation = 'portrait'
        elif orientation_id == SDL_ORIENTATION_PORTRAIT_FLIPPED:
            self.orientation = 'portrait_flipped'
        else:
            self.orientation = 'unknown'
        SDL_GetDisplayUsableBounds(display_id, bounds)
        self.usable_bounds = (bounds.x, bounds.y, bounds.w, bounds.h)

    def get_closest_mode(self, mode: DisplayMode) -> DisplayMode:
        display_mode = SDL_DisplayMode(mode.format.pixel_format, mode.w, mode.h, mode.refresh_rate)
        closest_mode = SDL_DisplayMode()
        SDL_GetClosestDisplayMode(self.display_id, display_mode, closest_mode)
        return DisplayMode(closest_mode)


class DisplaysManager:
    def __init__(self, app: any) -> None:
        self.destroyed = True
        self.app = app
        self.displays = [Display(i, app) for i in range(SDL_GetNumVideoDisplays())]
        self.destroyed = False

    def get_display_rect(self, given_rect: any) -> Display:
        screen_rect = SDL_Rect(int(given_rect[0]), int(given_rect[2]), int(given_rect[2]), int(given_rect[3]))
        return self.displays[SDL_GetRectDisplayIndex(screen_rect)]

    def get_display_point(self, point: any) -> Display:
        screen_point = SDL_Point(int(point[0]), int(point[1]))
        return self.displays[SDL_GetPointDisplayIndex(screen_point)]

    def get_window_display(self, window: any) -> Display:
        return self.displays[SDL_GetWindowDisplayIndex(window.window)]

    def get_window_display_mode(self, window: any) -> DisplayMode:
        dm_ptr = SDL_DisplayMode()
        SDL_GetWindowDisplayMode(window.window, dm_ptr)
        return DisplayMode(dm_ptr, self.app)

    @staticmethod
    def set_window_display_mode(window: any, display_mode: DisplayMode) -> None:
        SDL_SetWindowDisplayMode(window.window, display_mode.as_dm())

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        del self.app
        self.destroyed = True
        return False

    def __del__(self) -> None:
        self.destroy()


class Backend:
    def __init__(self, backend_id: int, app: any) -> None:
        self.backend_id = backend_id
        renderer_info = SDL_RendererInfo()
        SDL_GetRenderDriverInfo(backend_id, renderer_info)
        self.name = app.bts(renderer_info.name)
        self.is_software = bool(renderer_info.flags & SDL_RENDERER_SOFTWARE)
        self.is_accelerated = bool(renderer_info.flags & SDL_RENDERER_ACCELERATED)
        self.can_vsync = bool(renderer_info.flags & SDL_RENDERER_PRESENTVSYNC)
        self.can_target_texture = bool(renderer_info.flags & SDL_RENDERER_TARGETTEXTURE)
        self.texture_formats = []
        for format_id in range(renderer_info.num_texture_formats):
            self.texture_formats.append(PixelFormat(renderer_info.texture_formats[format_id], app))
        self.max_texture_width = renderer_info.max_texture_width
        self.max_texture_height = renderer_info.max_texture_height


class BackendManager:
    def __init__(self, app: any) -> None:
        self.drivers = [app.bts(SDL_GetVideoDriver(x)) for x in range(SDL_GetNumVideoDrivers())]
        self.current_driver = app.bts(SDL_GetCurrentVideoDriver())
        self.prefer_order = ['opengl', 'opengles2', 'opengles', 'software']
        if sys.platform == 'win32':
            win_ver = sys.getwindowsversion().major
            self.prefer_order.insert(0, 'direct3d')
            if win_ver >= 6:
                self.prefer_order.insert(0, 'direct3d11')
            if win_ver >= 10:
                self.prefer_order.insert(0, 'direct3d12')
        self.backends = [Backend(x, app) for x in range(SDL_GetNumRenderDrivers())]

    def get_by_name(self, backend_name: str) -> Backend:
        return self.backends[max(self.get_by_name_id(backend_name), 0)]

    def get_by_name_id(self, backend_name: str) -> int:
        for backend in self.backends:
            if backend.name == backend_name:
                return backend.backend_id
        return -1

    def get_best(self) -> Backend:
        return self.backends[max(self.get_best_id(), 0)]

    def get_best_id(self) -> int:
        for order in self.prefer_order:
            for backend in self.backends:
                if backend.name == order:
                    return backend.backend_id
        return -1



import ctypes
from .events import DropEvent, TouchFingerEvent, KeyboardEvent, MouseMotionEvent, MouseButtonEvent, MouseWheelEvent,\
    TextEditingEvent, TextInputEvent, WindowEvent, default_window_id
from .video import DisplayMode, PixelFormat
from .surface import Surface
from sdl2 import *

try:
    SDL_WINDOWEVENT_ICCPROF_CHANGED = SDL_WINDOWEVENT_ICCPROF_CHANGED
except NameError:
    SDL_WINDOWEVENT_ICCPROF_CHANGED = 17
try:
    SDL_WINDOWEVENT_DISPLAY_CHANGED = SDL_WINDOWEVENT_DISPLAY_CHANGED
except NameError:
    SDL_WINDOWEVENT_DISPLAY_CHANGED = 18


class Window:
    def __init__(
            self,
            app: any,
            pos: any = None,
            size: any = (640, 480),
            skip_taskbar: bool = False,
            window_type: str = None
    ) -> None:
        self.destroyed = True
        self.app = app
        self.x, self.y = pos or (SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED)
        self.w, self.h = size
        self.window = SDL_CreateWindow(
            app.stb('Good Window'),
            self.x, self.y,
            self.w, self.h,
            SDL_WINDOW_ALLOW_HIGHDPI | SDL_WINDOW_HIDDEN | (SDL_WINDOW_UTILITY if window_type == 'utility' else (
                SDL_WINDOW_TOOLTIP if window_type == 'tooltip' else (
                    SDL_WINDOW_POPUP_MENU if window_type == 'popup_menu' else 0
                )
            )) | (SDL_WINDOW_SKIP_TASKBAR if skip_taskbar else 0)
        )
        if not self.window:
            app.raise_error()
        self.mouse_rect = None
        self.shown = False
        self.borderless = False
        self.maximized = False
        self.minimized = False
        self.resizable = False
        self.always_on_top = False
        self.input_focused = False
        self.mouse_focused = False
        self.emulate_mouse_with_touch = False
        self.opacity = 1.0
        self.brightness = 1.0
        self.display_id = SDL_GetWindowDisplayIndex(self.window)
        self.min_size = (0, 0)
        self.max_size = (0, 0)
        self.event_map = {
            SDL_WINDOWEVENT_SHOWN: self.on_show,
            SDL_WINDOWEVENT_HIDDEN: self.on_hide,
            SDL_WINDOWEVENT_EXPOSED: self.on_expose,
            SDL_WINDOWEVENT_MOVED: self.on_move,
            SDL_WINDOWEVENT_RESIZED: self.on_resize,
            SDL_WINDOWEVENT_SIZE_CHANGED: self.on_size_change,
            SDL_WINDOWEVENT_MINIMIZED: self.on_minimize,
            SDL_WINDOWEVENT_MAXIMIZED: self.on_maximize,
            SDL_WINDOWEVENT_RESTORED: self.on_restore,
            SDL_WINDOWEVENT_ENTER: self.on_enter,
            SDL_WINDOWEVENT_LEAVE: self.on_leave,
            SDL_WINDOWEVENT_FOCUS_GAINED: self.on_focus_gain,
            SDL_WINDOWEVENT_FOCUS_LOST: self.on_focus_lose,
            SDL_WINDOWEVENT_CLOSE: self.on_close,
            SDL_WINDOWEVENT_TAKE_FOCUS: self.on_take_focus,
            SDL_WINDOWEVENT_ICCPROF_CHANGED: self.on_icc_profile_change,
            SDL_WINDOWEVENT_DISPLAY_CHANGED: self.on_display_change
        }
        borders = [ctypes.c_int(0), ctypes.c_int(0), ctypes.c_int(0), ctypes.c_int(0)]
        SDL_GetWindowBordersSize(self.window, *borders)
        self.borders_size = (borders[0].value, borders[1].value, borders[2].value, borders[3].value)
        self.pixel_format = PixelFormat(SDL_GetWindowPixelFormat(self.window), app)
        self.update_pos()
        self.update_size()
        self.display_mode = self.get_display_mode()
        self.id = SDL_GetWindowID(self.window)
        default_window_id[0] = self.id
        app.windows[self.id] = self
        self.destroyed = False
        # TODO:
        #  gamma ramp, hit test (maybe no?) and other spec. things

    @staticmethod
    def get_mouse_buttons() -> tuple:
        state = SDL_GetMouseState(None, None)
        return (
            bool(state & SDL_BUTTON_LMASK),
            bool(state & SDL_BUTTON_MMASK),
            bool(state & SDL_BUTTON_RMASK),
            bool(state & SDL_BUTTON_X1MASK),
            bool(state & SDL_BUTTON_X2MASK)
        )

    @staticmethod
    def get_mouse_pos() -> tuple:
        x_ptr, y_ptr = ctypes.c_int(), ctypes.c_int()
        SDL_GetMouseState(x_ptr, y_ptr)
        return x_ptr.value, y_ptr.value

    def set_mouse_pos(self, pos: any) -> None:
        SDL_WarpMouseInWindow(self.window, int(pos[0]), int(pos[1]))

    @staticmethod
    def screen_saver(enabled: bool) -> None:
        (SDL_EnableScreenSaver if enabled else SDL_DisableScreenSaver)()

    @staticmethod
    def is_screen_saver_enabled() -> bool:
        return bool(SDL_IsScreenSaverEnabled())

    def flash(self) -> None:
        SDL_FlashWindow(self.window)

    def focus_input(self) -> None:
        SDL_SetWindowInputFocus(self.window)

    def make_modal(self, parent_window: any) -> None:
        SDL_SetWindowModalFor(self.window, parent_window.window)

    def set_mouse_rect(self, mouse_rect: any = None) -> None:
        SDL_SetWindowMouseRect(
            self.window,
            mouse_rect and SDL_Rect(int(mouse_rect[0]), int(mouse_rect[1]), int(mouse_rect[2]), int(mouse_rect[3]))
        )
        self.mouse_rect = mouse_rect

    def set_mouse_grab(self, enabled: bool) -> None:
        SDL_SetWindowMouseGrab(self.window, enabled)

    def set_keyboard_grab(self, enabled: bool) -> None:
        SDL_SetWindowKeyboardGrab(self.window, enabled)

    def get_icc_profile(self) -> bytes:
        size_ptr = ctypes.c_size_t()
        data_ptr = SDL_GetWindowICCProfile(self.window, size_ptr)
        data_buf = (ctypes.c_uint8 * size_ptr.value).from_address(data_ptr)
        return bytes(data_buf[:size_ptr.value])  # noqa

    def set_title(self, title: str) -> None:
        SDL_SetWindowTitle(self.window, self.app.stb(title))

    def set_grab(self, grabbed: bool) -> None:
        SDL_SetWindowGrab(self.window, grabbed)

    def get_grab(self) -> bool:
        return SDL_GetWindowGrab(self.window) == SDL_TRUE

    def get_display_mode(self) -> DisplayMode:
        display_mode = SDL_DisplayMode()
        SDL_GetWindowDisplayMode(self.window, display_mode)
        return DisplayMode(display_mode, self.app)

    def set_display_mode(self, display_mode: DisplayMode) -> None:
        SDL_SetWindowDisplayMode(self.window, display_mode.as_dm())
        self.display_mode = self.get_display_mode()

    def show(self) -> None:
        SDL_ShowWindow(self.window)
        self.update_flags()

    def hide(self) -> None:
        SDL_HideWindow(self.window)
        self.update_flags()

    def restore(self) -> None:
        SDL_RestoreWindow(self.window)
        self.update_flags()

    def maximize(self) -> None:
        SDL_MaximizeWindow(self.window)
        self.update_flags()

    def minimize(self) -> None:
        SDL_MinimizeWindow(self.window)
        self.update_flags()

    def get_hwnd(self) -> int:
        wm_info = SDL_SysWMinfo()
        SDL_VERSION(wm_info.version)
        SDL_GetWindowWMInfo(self.window, wm_info)
        return wm_info.info.win.window

    def update_flags(self) -> None:
        flags = SDL_GetWindowFlags(self.window)
        self.maximized = bool(flags & SDL_WINDOW_MAXIMIZED)
        self.minimized = bool(flags & SDL_WINDOW_MINIMIZED)
        self.shown = bool(flags & SDL_WINDOW_SHOWN)
        self.input_focused = bool(flags & SDL_WINDOW_INPUT_FOCUS)
        self.mouse_focused = bool(flags & SDL_WINDOW_MOUSE_FOCUS)
        self.brightness = SDL_GetWindowBrightness(self.window)

    def update_pos(self) -> None:
        x_ptr, y_ptr = ctypes.c_int(), ctypes.c_int()
        SDL_GetWindowPosition(self.window, x_ptr, y_ptr)
        self.x, self.y = x_ptr.value, y_ptr.value

    def update_size(self) -> None:
        w_ptr, h_ptr = ctypes.c_int(), ctypes.c_int()
        SDL_GetWindowSize(self.window, w_ptr, h_ptr)
        self.w, self.h = w_ptr.value, h_ptr.value

    def set_pos(self, pos: any) -> None:
        if not pos:
            pos = (SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED)
        SDL_SetWindowPosition(self.window, pos[0], pos[1])
        self.update_pos()

    def set_size(self, size: any) -> None:
        SDL_SetWindowSize(self.window, size[0], size[1])
        self.update_size()

    def set_min_size(self, w: int, h: int) -> None:
        self.min_size = (w, h)
        SDL_SetWindowMinimumSize(self.window, w, h)
        self.update_size()

    def set_max_size(self, w: int, h: int) -> None:
        self.max_size = (w, h)
        SDL_SetWindowMaximumSize(self.window, w, h)
        self.update_size()

    def set_window_mode(self, mode: str = 'windowed') -> None:
        full_screen = 0
        if mode == 'full_screen':
            full_screen |= SDL_WINDOW_FULLSCREEN
        elif mode == 'desktop':
            full_screen |= SDL_WINDOW_FULLSCREEN_DESKTOP
        elif not mode == 'windowed':
            raise RuntimeError('Only full_screen, desktop and windowed modes are allowed!')
        SDL_SetWindowFullscreen(self.window, full_screen)

    def set_borderless(self, borderless: bool) -> None:
        self.borderless = borderless
        SDL_SetWindowBordered(self.window, not borderless)

    def set_resizable(self, resizable: bool) -> None:
        self.resizable = resizable
        SDL_SetWindowResizable(self.window, resizable)

    def set_always_on_top(self, always_on_top: bool) -> None:
        self.always_on_top = always_on_top
        SDL_SetWindowAlwaysOnTop(self.window, always_on_top)

    def set_opacity(self, opacity: float) -> None:
        self.opacity = opacity
        SDL_SetWindowOpacity(self.window, opacity)

    def set_brightness(self, brightness: float) -> None:
        SDL_SetWindowBrightness(self.window, brightness)
        self.update_size()

    def show_message_box(
            self,
            title: str,
            text: str,
            icon_type: str = None
    ) -> None:
        return self.app.show_message_box(title, text, icon_type, self)

    def is_screen_keyboard_shown(self) -> bool:
        return bool(SDL_IsScreenKeyboardShown(self.window))

    def set_icon(self, surf: Surface) -> None:
        SDL_SetWindowIcon(self.window, surf.surface)

    def raise_self(self) -> None:
        SDL_RaiseWindow(self.window)

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        self.destroyed = True
        SDL_DestroyWindow(self.window)
        del self.app.windows[self.id]
        del self.app
        return False

    def on_drop_file(self, event: DropEvent) -> None:
        pass

    def on_drop_text(self, event: DropEvent) -> None:
        pass

    def on_drop_begin(self, event: DropEvent) -> None:
        pass

    def on_drop_complete(self, event: DropEvent) -> None:
        pass

    def on_finger_move(self, event: TouchFingerEvent) -> None:
        pass

    def on_finger_down(self, event: TouchFingerEvent) -> None:
        pass

    def on_finger_up(self, event: TouchFingerEvent) -> None:
        pass

    def on_key_down(self, event: KeyboardEvent) -> None:
        pass

    def on_key_up(self, event: KeyboardEvent) -> None:
        pass

    def on_text_edit(self, event: TextEditingEvent) -> None:
        pass

    def on_text_edit_ext(self, event: TextEditingEvent) -> None:
        pass

    def on_mouse_move(self, event: MouseMotionEvent) -> None:
        pass

    def on_mouse_down(self, event: MouseButtonEvent) -> None:
        pass

    def on_mouse_up(self, event: MouseButtonEvent) -> None:
        pass

    def on_mouse_wheel(self, event: MouseWheelEvent) -> None:
        pass

    def on_text_input(self, event: TextInputEvent) -> None:
        pass

    def on_show(self, event: WindowEvent) -> None:
        pass

    def on_hide(self, event: WindowEvent) -> None:
        self.update_flags()

    def on_expose(self, event: WindowEvent) -> None:
        self.update_flags()

    def on_move(self, event: WindowEvent) -> None:
        self.x, self.y = event.data1, event.data2

    def on_resize(self, event: WindowEvent) -> None:
        self.w, self.h = event.data1, event.data2

    def on_size_change(self, event: WindowEvent) -> None:
        self.w, self.h = event.data1, event.data2

    def on_minimize(self, event: WindowEvent) -> None:
        self.update_flags()

    def on_maximize(self, event: WindowEvent) -> None:
        self.update_flags()

    def on_restore(self, event: WindowEvent) -> None:
        self.update_flags()

    def on_enter(self, event: WindowEvent) -> None:
        self.update_flags()

    def on_leave(self, event: WindowEvent) -> None:
        self.update_flags()

    def on_focus_gain(self, event: WindowEvent) -> None:
        self.update_flags()

    def on_focus_lose(self, event: WindowEvent) -> None:
        self.update_flags()

    def on_close(self, event: WindowEvent) -> None:
        pass

    def on_take_focus(self, event: WindowEvent) -> None:
        self.update_flags()

    def on_icc_profile_change(self, event: WindowEvent) -> None:
        pass

    def on_display_change(self, event: WindowEvent) -> None:
        self.display_mode = self.get_display_mode()

    def __del__(self) -> None:
        self.destroy()



from sdl2 import *
from .video import Display

try:
    SDL_TOUCH_MOUSEID = SDL_TOUCH_MOUSEID
except NameError:
    SDL_TOUCH_MOUSEID = 2 ** 32 - 1

default_window_id = [1]


class CommonEvent:
    def __init__(self, event: SDL_Event) -> None:
        self.event = event
        self.timestamp = event.timestamp
        self.type = event.type


class QuitEvent(CommonEvent):
    pass


class JoyAxisEvent(CommonEvent):
    def __init__(self, event: SDL_Event) -> None:
        super().__init__(event)
        self.instance_id = event.which
        self.axis = event.axis
        self.value = event.value


class JoyBallEvent(CommonEvent):
    def __init__(self, event: SDL_Event) -> None:
        super().__init__(event)
        self.instance_id = event.which
        self.ball = event.ball
        self.rel = (event.xrel, event.yrel)


class JoyButtonEvent(CommonEvent):
    def __init__(self, event: SDL_Event) -> None:
        super().__init__(event)
        self.instance_id = event.which
        self.button = event.button
        self.state = 'pressed' if event.state == SDL_PRESSED else 'released'


class JoyBatteryEvent(CommonEvent):
    def __init__(self, event: SDL_Event) -> None:
        super().__init__(event)
        self.instance_id = event.which
        self.button = event.button
        self.level = {
            SDL_JOYSTICK_POWER_UNKNOWN: 'unknown',
            SDL_JOYSTICK_POWER_EMPTY: 'empty',
            SDL_JOYSTICK_POWER_LOW: 'low',
            SDL_JOYSTICK_POWER_MEDIUM: 'medium',
            SDL_JOYSTICK_POWER_FULL: 'full',
            SDL_JOYSTICK_POWER_WIRED: 'wired',
            SDL_JOYSTICK_POWER_MAX: 'max'
        }.get(event.level)


class JoyDeviceEvent(CommonEvent):
    def __init__(self, event: SDL_Event) -> None:
        super().__init__(event)
        if self.type == SDL_JOYDEVICEADDED:
            self.index = event.which
            self.instance_id = 0
        else:
            self.index = 0
            self.instance_id = event.which


class JoyHatEvent(CommonEvent):
    def __init__(self, event: SDL_Event) -> None:
        super().__init__(event)
        self.instance_id = event.which
        self.hat = event.hat
        self.state = {
            SDL_JOYSTICK_POWER_UNKNOWN: 'unknown',
            SDL_JOYSTICK_POWER_EMPTY: 'empty',
            SDL_JOYSTICK_POWER_LOW: 'low',
            SDL_JOYSTICK_POWER_MEDIUM: 'medium',
            SDL_JOYSTICK_POWER_FULL: 'full',
            SDL_JOYSTICK_POWER_WIRED: 'wired',
            SDL_JOYSTICK_POWER_MAX: 'max'
        }.get(event.value)


class ControllerAxisEvent(CommonEvent):
    def __init__(self, event: SDL_Event) -> None:
        super().__init__(event)
        self.instance_id = event.which
        self.axis = {
            SDL_CONTROLLER_AXIS_INVALID: 'invalid',
            SDL_CONTROLLER_AXIS_LEFTX: 'left_x',
            SDL_CONTROLLER_AXIS_LEFTY: 'left_y',
            SDL_CONTROLLER_AXIS_RIGHTX: 'right_x',
            SDL_CONTROLLER_AXIS_RIGHTY: 'right_y',
            SDL_CONTROLLER_AXIS_TRIGGERLEFT: 'trigger_left',
            SDL_CONTROLLER_AXIS_TRIGGERRIGHT: 'trigger_right',
            SDL_CONTROLLER_AXIS_MAX: 'max'
        }.get(event.axis)
        self.value = event.value


class ControllerButtonEvent(CommonEvent):
    def __init__(self, event: SDL_Event) -> None:
        super().__init__(event)
        self.instance_id = event.which
        self.button = {
            SDL_CONTROLLER_BUTTON_INVALID: 'invalid',
            SDL_CONTROLLER_BUTTON_A: 'a',
            SDL_CONTROLLER_BUTTON_B: 'b',
            SDL_CONTROLLER_BUTTON_X: 'x',
            SDL_CONTROLLER_BUTTON_Y: 'y',
            SDL_CONTROLLER_BUTTON_BACK: 'back',
            SDL_CONTROLLER_BUTTON_GUIDE: 'guide',
            SDL_CONTROLLER_BUTTON_START: 'start',
            SDL_CONTROLLER_BUTTON_LEFTSTICK: 'left_stick',
            SDL_CONTROLLER_BUTTON_RIGHTSTICK: 'right_stick',
            SDL_CONTROLLER_BUTTON_LEFTSHOULDER: 'left_shoulder',
            SDL_CONTROLLER_BUTTON_RIGHTSHOULDER: 'right_shoulder',
            SDL_CONTROLLER_BUTTON_DPAD_UP: 'd_pad_up',
            SDL_CONTROLLER_BUTTON_DPAD_DOWN: 'd_pad_down',
            SDL_CONTROLLER_BUTTON_DPAD_LEFT: 'd_pad_left',
            SDL_CONTROLLER_BUTTON_DPAD_RIGHT: 'd_pad_right',
            SDL_CONTROLLER_BUTTON_MISC1: 'misc1',
            SDL_CONTROLLER_BUTTON_PADDLE1: 'paddle1',
            SDL_CONTROLLER_BUTTON_PADDLE2: 'paddle2',
            SDL_CONTROLLER_BUTTON_PADDLE3: 'paddle2',
            SDL_CONTROLLER_BUTTON_PADDLE4: 'paddle3',
            SDL_CONTROLLER_BUTTON_TOUCHPAD: 'touchpad',
            SDL_CONTROLLER_BUTTON_MAX: 'max'
        }.get(event.button)
        self.state = 'pressed' if event.state == SDL_PRESSED else 'released'


class ControllerDeviceEvent(CommonEvent):
    def __init__(self, event: SDL_Event) -> None:
        super().__init__(event)
        if self.type == SDL_JOYDEVICEADDED:
            self.index = event.which
            self.instance_id = 0
        else:
            self.index = 0
            self.instance_id = event.which


class ControllerTouchpadEvent(CommonEvent):
    def __init__(self, event: SDL_Event) -> None:
        super().__init__(event)
        self.instance_id = event.which
        self.touchpad = event.touchpad
        self.finger = event.finger
        self.pos = (event.x, event.y)
        self.pressure = event.pressure


class ControllerSensorEvent(CommonEvent):
    def __init__(self, event: SDL_Event) -> None:
        super().__init__(event)
        self.instance_id = event.which
        self.sensor = event.sensor
        self.pos = (event.x, event.y)
        self.data = (event.data[0], event.data[1], event.data[2])
        self.time = event.timestamp_us / 1000000


class AudioDeviceEvent(CommonEvent):
    def __init__(self, event: SDL_Event) -> None:
        super().__init__(event)
        if self.type == SDL_AUDIODEVICEADDED:
            self.which = event.which
            self.event = 'add'
        else:
            self.which = event.which
            self.event = 'remove'
        if event.iscapture == 0:
            self.device_type = 'playback'
        else:
            self.device_type = 'recording'


class DropEvent(CommonEvent):
    def __init__(self, event: SDL_Event, app: any) -> None:
        super().__init__(event)
        if self.type == SDL_DROPFILE or self.type == SDL_DROPTEXT:
            self.file = app.bts(event.file)
            self.drop_type = 'file' if self.type == SDL_DROPFILE else 'text'
            self.drop_state = None
        else:
            self.file = None
            self.drop_type = None
            self.drop_state = 'begin' if self.type == SDL_DROPBEGIN else 'complete'
        self.window_id = event.windowID if hasattr(event, 'windowID') else default_window_id[0]
        self.window = app.windows[self.window_id]


class TouchFingerEvent(CommonEvent):
    def __init__(self, event: SDL_Event, app: any) -> None:
        super().__init__(event)
        if self.type == SDL_FINGERMOTION:
            self.event = 'move'
        else:
            self.event = 'down' if self.type == SDL_FINGERDOWN else 'up'
        self.touch_id = event.touchId
        self.finger_id = event.fingerId
        self.pos = (event.x, event.y)
        self.d_pos = (event.dx, event.dy)
        self.pressure = event.pressure
        self.window_id = event.windowID if hasattr(event, 'windowID') else default_window_id[0]
        self.window = app.windows[self.window_id]


class KeyboardEvent(CommonEvent):
    def __init__(self, event: SDL_Event, app: any) -> None:
        super().__init__(event)
        self.event = 'down' if self.type == SDL_KEYDOWN else 'up'
        self.state = 'pressed' if event.state == SDL_PRESSED else 'released'
        self.repeat = event.repeat
        if self.repeat:
            self.event = 'hold'
        self.scancode_id = event.keysym.scancode
        self.scancode = app.bts(SDL_GetScancodeName(self.scancode_id))
        self.sym_id = event.keysym.sym
        self.sym = app.bts(SDL_GetKeyName(self.sym_id))
        self.mod = event.keysym.mod
        self.window_id = event.windowID if hasattr(event, 'windowID') else default_window_id[0]
        self.window = app.windows[self.window_id]


class MouseMotionEvent(CommonEvent):
    def __init__(self, event: SDL_Event, app: any) -> None:
        super().__init__(event)
        self.which = event.which
        self.emulated = self.which == SDL_TOUCH_MOUSEID
        self.state_num = event.state
        self.state = (
            bool(event.state & SDL_BUTTON_LMASK),
            bool(event.state & SDL_BUTTON_MMASK),
            bool(event.state & SDL_BUTTON_RMASK),
            bool(event.state & SDL_BUTTON_X1MASK),
            bool(event.state & SDL_BUTTON_X2MASK)
        )
        self.pos = (event.x, event.y)
        self.rel = (event.xrel, event.yrel)
        self.window_id = event.windowID if hasattr(event, 'windowID') else default_window_id[0]
        self.window = app.windows[self.window_id]


class MouseButtonEvent(CommonEvent):
    def __init__(self, event: SDL_Event, app: any) -> None:
        super().__init__(event)
        self.button_event = 'down' if event.type == SDL_MOUSEBUTTONDOWN else 'up'
        self.which = event.which
        self.emulated = self.which == SDL_TOUCH_MOUSEID
        self.button_id = event.button
        self.button = {
            SDL_BUTTON_LEFT: 0,
            SDL_BUTTON_MIDDLE: 1,
            SDL_BUTTON_RIGHT: 2,
            SDL_BUTTON_X1: 3,
            SDL_BUTTON_X2: 4
        }.get(self.button_id)
        self.state = 'pressed' if event.state == SDL_PRESSED else 'released'
        self.clicks = event.clicks
        self.pos = (event.x, event.y)
        self.window_id = event.windowID if hasattr(event, 'windowID') else default_window_id[0]
        self.window = app.windows[self.window_id]


class MouseWheelEvent(CommonEvent):
    def __init__(self, event: SDL_Event, app: any) -> None:
        super().__init__(event)
        self.which = event.which
        self.emulated = self.which == SDL_TOUCH_MOUSEID
        self.pos = (event.x, event.y)
        self.direction = 'flipped' if event.direction == SDL_MOUSEWHEEL_FLIPPED else 'normal'
        self.precise_x = event.preciseX
        self.precise_y = event.preciseY
        self.window_id = event.windowID if hasattr(event, 'windowID') else default_window_id[0]
        self.window = app.windows[self.window_id]


class TextEditingEvent(CommonEvent):
    def __init__(self, event: SDL_Event, app: any) -> None:
        super().__init__(event)
        self.text = app.bts(event.text, 'utf-8')
        self.start = event.start
        self.length = event.length
        self.window_id = event.windowID if hasattr(event, 'windowID') else default_window_id[0]
        self.window = app.windows[self.window_id]


class TextInputEvent(CommonEvent):
    def __init__(self, event: SDL_Event, app: any) -> None:
        super().__init__(event)
        self.text = app.bts(event.text, 'utf-8')
        self.window_id = event.windowID if hasattr(event, 'windowID') else default_window_id[0]
        self.window = app.windows[self.window_id]


class DisplayEvent(CommonEvent):
    def __init__(self, event: SDL_Event, app: any) -> None:
        super().__init__(event)
        self.display = Display(event.display, app)
        if event.event == SDL_DISPLAYEVENT_ORIENTATION:
            self.event = 'orientation_change'
        elif event.event == SDL_DISPLAYEVENT_CONNECTED:
            self.event = 'connect'
        elif event.event == SDL_DISPLAYEVENT_DISCONNECTED:
            self.event = 'disconnect'
        else:
            self.event = None
        self.data = event.data1


class WindowEvent(CommonEvent):
    def __init__(self, event: SDL_Event, app: any) -> None:
        super().__init__(event)
        self.data1 = event.data1
        self.data2 = event.data2
        self.event = event.event
        self.padding1 = event.padding1
        self.padding2 = event.padding2
        self.padding3 = event.padding3
        self.window_id = event.windowID if hasattr(event, 'windowID') else default_window_id[0]
        self.window = app.windows[self.window_id]



from sdl2 import *


class Finger:
    def __init__(self, finger: SDL_Finger) -> None:
        self.id = finger.id
        self.x, self.y = finger.x, finger.y
        self.pressure = finger.pressure


class TouchDevice:
    def __init__(self, app: any, index: int) -> None:
        self.index = index
        self.id = SDL_GetTouchDevice(index)
        try:
            self.name = app.bts(SDL_GetTouchName(index))
        except NameError:
            self.name = ''
        try:
            tp = SDL_GetTouchDeviceType(self.id)
            if tp == SDL_TOUCH_DEVICE_DIRECT:
                self.type = 'direct'
            elif SDL_TOUCH_DEVICE_INDIRECT_ABSOLUTE:
                self.type = 'indirect_absolute'
            elif SDL_TOUCH_DEVICE_INDIRECT_RELATIVE:
                self.type = 'indirect_relative'
            else:
                self.type = 'invalid'
        except NameError:
            self.type = 'unknown'
        self.num_fingers = SDL_GetNumTouchFingers(self.id)

    def get_finger(self, finger_id: int) -> Finger:
        return Finger(SDL_GetTouchFinger(self.id, finger_id))

    def get_all_fingers(self) -> tuple:
        return tuple(self.get_finger(_x) for _x in range(self.num_fingers))



import ctypes
from sdl2 import *


class GameController:
    def __init__(self, app: any, index: int) -> None:
        self.app = app
        self.destroyed = True
        self.index = index
        self.name = app.bts(SDL_GameControllerNameForIndex(index))
        try:
            self.path = app.bts(SDL_GameControllerPathForIndex(index))
        except NameError:
            self.path = ''
        try:
            self.type = {
                SDL_CONTROLLER_TYPE_UNKNOWN: 'unknown',
                SDL_CONTROLLER_TYPE_XBOX360: 'xbox_360',
                SDL_CONTROLLER_TYPE_XBOXONE: 'xbox_one',
                SDL_CONTROLLER_TYPE_PS3: 'ps3',
                SDL_CONTROLLER_TYPE_PS4: 'ps4',
                SDL_CONTROLLER_TYPE_NINTENDO_SWITCH_PRO: 'nintendo_switch_pro',
                SDL_CONTROLLER_TYPE_VIRTUAL: 'virtual',
                SDL_CONTROLLER_TYPE_PS5: 'ps5',
                SDL_CONTROLLER_TYPE_AMAZON_LUNA: 'amazon_luna',
                SDL_CONTROLLER_TYPE_GOOGLE_STADIA: 'google_stadia',
                SDL_CONTROLLER_TYPE_NVIDIA_SHIELD: 'nvidia_shield',
                SDL_CONTROLLER_TYPE_NINTENDO_SWITCH_JOYCON_LEFT: 'nintendo_switch_joy_con_left',
                SDL_CONTROLLER_TYPE_NINTENDO_SWITCH_JOYCON_RIGHT: 'nintendo_switch_joy_con_right',
                SDL_CONTROLLER_TYPE_NINTENDO_SWITCH_JOYCON_PAIR: 'nintendo_switch_joy_con_right'
            }.get(SDL_GameControllerTypeForIndex(index))
        except NameError:
            self.type = 'unknown'
        self.opened = False
        self.player_index = -1
        self.vendor = 0
        self.product = 0
        self.product_version = 0
        self.firmware_version = 0
        self.serial = b''
        self.controller = SDL_GameController()
        self.destroyed = False
        # TODO:
        #  finish

    def add_mappings_from_file(self, path: str) -> None:
        SDL_GameControllerAddMappingsFromFile(self.app.stb(path))

    def add_mapping(self, mapping: str) -> None:
        SDL_GameControllerAddMapping(self.app.stb(mapping))

    def is_attached(self) -> bool:
        return bool(SDL_GameControllerGetAttached(self.controller))

    def set_player_index(self, player_index: int = -1) -> None:
        self.player_index = player_index
        SDL_GameControllerSetPlayerIndex(self.controller, player_index)

    def open(self) -> None:
        self.controller = SDL_GameControllerOpen(self.index)
        self.vendor = SDL_GameControllerGetVendor(self.controller)
        self.product = SDL_GameControllerGetProduct(self.controller)
        self.product_version = SDL_GameControllerGetProductVersion(self.controller)
        try:
            self.firmware_version = SDL_GameControllerGetFirmwareVersion(self.controller)
        except NameError:
            pass
        try:
            self.serial = SDL_GameControllerGetSerial(self.controller)
        except NameError:
            pass
        try:
            self.player_index = SDL_GameControllerGetPlayerIndex(self.controller)
        except NameError:
            return

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        if self.opened:
            self.opened = False
            SDL_GameControllerClose(self.controller)
        del self.app
        self.destroyed = True
        return False

    def __del__(self) -> None:
        self.destroy()



import os
import sys
import ctypes
from .sdl import sdl_dir
from sdl2 import *

if os.getenv('GG_ENABLE_NUMBA') == '1':
    try:
        import numba
    except Exception as _err:
        from . import numba
        print(f'Failed to import numba [{_err}]. JIT for Math will be disabled!')
else:
    from . import numba


class Math:
    def __init__(self) -> None:
        self.destroyed = True
        if 'SDL_HasIntersectionF' not in sdl_dir:
            self.has_intersection = self.has_intersection_i
            self.intersect = self.intersect_i
            self.union = self.union_i
            self.intersect_rect_and_line = self.intersect_rect_and_line_i
            self.enclose_points = self.enclose_points_i
        self.destroyed = False
        # TODO
        #  add more functions

    @staticmethod
    def enclose_points(r: any, points: any) -> tuple:
        result = SDL_FRect()
        SDL_EncloseFPoints(
            (SDL_FPoint * len(points))(*(SDL_FPoint(point[0], point[1]) for point in points)),
            len(points),
            SDL_FRect(r[0], r[1], r[2], r[3]),
            result
        )
        return result.x, result.y, result.w, result.h

    @staticmethod
    def enclose_points_i(r: any, points: any) -> tuple:
        result = SDL_Rect()
        SDL_EnclosePoints(
            (SDL_Point * len(points))(*(SDL_Point(int(point[0]), int(point[1])) for point in points)),
            len(points),
            SDL_Rect(int(r[0]), int(r[1]), int(r[2]), int(r[3])),
            result
        )
        return result.x, result.y, result.w, result.h

    @staticmethod
    def intersect_rect_and_line(r: any, start: any, end: any) -> tuple:
        rect_ptr = SDL_FRect(r[0], r[1], r[2], r[3])
        x1_ptr, y1_ptr = ctypes.c_float(start[0]), ctypes.c_float(start[1])
        x2_ptr, y2_ptr = ctypes.c_float(end[0]), ctypes.c_float(end[1])
        SDL_IntersectFRectAndLine(rect_ptr, x1_ptr, y1_ptr, x2_ptr, y2_ptr)
        return x1_ptr.value, y1_ptr.value, x2_ptr.value, y2_ptr.value

    @staticmethod
    def intersect_rect_and_line_i(r: any, start: any, end: any) -> tuple:
        rect_ptr = SDL_Rect(int(r[0]), int(r[1]), int(r[2]), int(r[3]))
        x1_ptr, y1_ptr = ctypes.c_int(int(start[0])), ctypes.c_int(int(start[1]))
        x2_ptr, y2_ptr = ctypes.c_int(int(end[0])), ctypes.c_int(int(end[1]))
        SDL_IntersectRectAndLine(rect_ptr, x1_ptr, y1_ptr, x2_ptr, y2_ptr)
        return x1_ptr.value, y1_ptr.value, x2_ptr.value, y2_ptr.value

    @staticmethod
    def union(a: any, b: any) -> any:
        result = SDL_FRect()
        SDL_UnionFRect(
            SDL_FRect(a[0], a[1], a[2], a[3]),
            SDL_FRect(b[0], b[1], b[2], b[3]),
            result
        )
        return result.x, result.y, result.w, result.h

    @staticmethod
    def union_i(a: any, b: any) -> any:
        result = SDL_Rect()
        SDL_UnionRect(
            SDL_Rect(int(a[0]), int(a[1]), int(a[2]), int(a[3])),
            SDL_Rect(int(b[0]), int(b[1]), int(b[2]), int(b[3])),
            result
        )
        return result.x, result.y, result.w, result.h

    @staticmethod
    def intersect(a: any, b: any) -> any:
        result = SDL_FRect()
        if not SDL_IntersectFRect(
            SDL_FRect(a[0], a[1], a[2], a[3]),
            SDL_FRect(b[0], b[1], b[2], b[3]),
            result
        ):
            return None
        return result.x, result.y, result.w, result.h

    @staticmethod
    def intersect_i(a: any, b: any) -> any:
        result = SDL_Rect()
        if not SDL_IntersectRect(
            SDL_Rect(int(a[0]), int(a[1]), int(a[2]), int(a[3])),
            SDL_Rect(int(b[0]), int(b[1]), int(b[2]), int(b[3])),
            result
        ):
            return None
        return result.x, result.y, result.w, result.h

    @staticmethod
    def has_intersection(a: any, b: any) -> bool:
        return bool(SDL_HasIntersectionF(
            SDL_FRect(a[0], a[1], a[2], a[3]),
            SDL_FRect(b[0], b[1], b[2], b[3])
        ))

    @staticmethod
    def has_intersection_i(a: any, b: any) -> bool:
        return bool(SDL_HasIntersection(
            SDL_Rect(int(a[0]), int(a[1]), int(a[2]), int(a[3])),
            SDL_Rect(int(b[0]), int(b[1]), int(b[2]), int(b[3]))
        ))

    @staticmethod
    @numba.jit(nopython=True, fastmath=True)
    def point_in_rect(r: any, p: any) -> bool:
        return r[0] <= p[0] < r[0] + r[2] and r[1] <= p[1] < r[1] + r[3]

    @staticmethod
    @numba.jit(nopython=True, fastmath=True)
    def rect_empty(r: any) -> bool:
        return r[0] <= 0 or r[1] <= 0

    @staticmethod
    @numba.jit(nopython=True, fastmath=True)
    def rect_equals(a: any, b: any, epsilon: float = sys.float_info.epsilon) -> bool:
        return abs(a[0] - b[0]) <= epsilon and abs(a[1] - b[1]) <= epsilon and\
            abs(a[2] - b[2]) <= epsilon and abs(a[3] - b[3]) <= epsilon

    @staticmethod
    @numba.jit(nopython=True, fastmath=True)
    def rect_equals_i(a: any, b: any) -> bool:
        return int(a[0]) == int(b[0]) and int(a[1]) == int(b[1]) and int(a[2]) == int(b[2]) and int(a[3]) == int(b[3])

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        self.has_intersection = None
        self.intersect = None
        self.union = None
        self.intersect_rect_and_line = None
        self.enclose_points = None
        self.destroyed = True
        return False



import ctypes
from .video import Backend, BackendManager
from .surface import Surface
from .texture import Texture
from .sdl import sdl_dir
from sdl2 import *

try:
    from sdl2.sdlimage import *
except:  # noqa
    pass
try:
    from sdl2.sdlgfx import *
except Exception as _err:
    from .sdlgfx import *
    print(f'Failed to import SDL2_gfx [{_err}]. Using extremely slow and unfinished software fallback!')


class Renderer:
    def __init__(
            self,
            window: any,
            backend: Backend = None,
            vsync: bool = False,
            force_int: bool = False
    ) -> None:
        self.destroyed = True
        self.app = window.app
        self.window = window
        self.backend = backend or BackendManager(self.app).get_best()
        self.vsync = vsync
        self.texture = None
        if force_int or 'SDL_RenderCopyF' not in sdl_dir:
            self.blit = self.blit_i
            self.blit_ex = self.blit_ex_i
            self.draw_point = self.draw_point_i
            self.draw_points = self.draw_points_i
            self.sdl_draw_rect = self.sdl_draw_rect_i
            self.sdl_draw_rects = self.sdl_draw_rects_i
            self.sdl_fill_rect = self.sdl_fill_rect_i
            self.sdl_fill_rects = self.sdl_fill_rects_i
            self.sdl_draw_line = self.sdl_draw_line_i
            self.sdl_draw_lines = self.sdl_draw_lines_i
        self.renderer = SDL_CreateRenderer(
            window.window,
            self.backend.backend_id,
            SDL_RENDERER_ACCELERATED | SDL_RENDERER_TARGETTEXTURE | (SDL_RENDERER_PRESENTVSYNC if vsync else 0)
        )
        self.render_target_supported = bool(SDL_RenderTargetSupported(self.renderer))
        self.destroyed = False
        # TODO:
        #  check out of bounds (check if this handled automatic by sdl)
        #  SDL_RenderReadPixels, SDL_RenderGeometry, SDL_RenderGeometryRaw
        #  Fix Scaling for SDL2_gfx

    def window_to_logical(self, pos: any) -> tuple:
        w_ptr, h_ptr = ctypes.c_float(), ctypes.c_float()
        SDL_RenderWindowToLogical(self.renderer, int(pos[0]), int(pos[1]), w_ptr, h_ptr)
        return w_ptr.value, h_ptr.value

    def logical_to_window(self, pos: any) -> tuple:
        w_ptr, h_ptr = ctypes.c_int(), ctypes.c_int()
        SDL_RenderLogicalToWindow(self.renderer, pos[0], pos[1], w_ptr, h_ptr)
        return w_ptr.value, h_ptr.value

    def gt_scale(self) -> tuple:
        scale_x_ptr, scale_y_ptr = ctypes.c_float(), ctypes.c_float()
        SDL_RenderGetScale(self.renderer, scale_x_ptr, scale_y_ptr)
        return scale_x_ptr.value, scale_y_ptr.value

    def set_scale(self, scale: any = (1.0, 1.0)) -> None:
        SDL_RenderSetScale(self.renderer, scale[0], scale[1])

    def is_clip_enabled(self) -> bool:
        return bool(SDL_RenderIsClipEnabled(self.renderer))

    def set_clip_rect(self, clip_rect: any = None) -> None:
        SDL_RenderSetClipRect(
            self.renderer,
            clip_rect and SDL_Rect(int(clip_rect[0]), int(clip_rect[1]), int(clip_rect[2]), int(clip_rect[3]))
        )

    def get_clip_rect(self) -> tuple:
        clip_rect_ptr = SDL_Rect()
        SDL_RenderGetClipRect(self.renderer, clip_rect_ptr)
        return clip_rect_ptr.x, clip_rect_ptr.y, clip_rect_ptr.w, clip_rect_ptr.h

    def set_viewport(self, viewport_rect: any = None) -> None:
        SDL_RenderSetViewport(
            self.renderer,
            viewport_rect and SDL_Rect(int(viewport_rect[0]), int(viewport_rect[1]),
                                       int(viewport_rect[2]), int(viewport_rect[3]))
        )

    def get_viewport(self) -> tuple:
        viewport_ptr = SDL_Rect()
        SDL_RenderGetViewport(self.renderer, viewport_ptr)
        return viewport_ptr.x, viewport_ptr.y, viewport_ptr.w, viewport_ptr.h

    def set_logical_size(self, size: any) -> None:
        SDL_RenderSetLogicalSize(self.renderer, int(size[0]), int(size[1]))

    def get_logical_size(self, size: any) -> tuple:
        w_ptr, h_ptr = ctypes.c_int(), ctypes.c_int()
        SDL_RenderGetLogicalSize(self.renderer, w_ptr, h_ptr)
        return w_ptr, h_ptr

    def set_target(self, target: any = None) -> None:
        SDL_SetRenderTarget(self.renderer, target and target.texture)

    def texture_from_file(self, path: str) -> Texture:
        texture = IMG_LoadTexture(self.renderer, self.app.stb(path))
        if not texture:
            self.app.raise_error(IMG_GetError)
        return Texture(texture, self)

    def blit(self, texture: Texture, src_rect: any = None, dst_rect: any = None) -> None:
        SDL_RenderCopyF(
            self.renderer,
            texture.texture,
            src_rect and SDL_FRect(src_rect[0], src_rect[1], src_rect[2], src_rect[3]),
            dst_rect and (SDL_FRect(dst_rect[0], dst_rect[1], dst_rect[2], dst_rect[3]) if len(dst_rect) > 2 else
                          SDL_FRect(dst_rect[0], dst_rect[1], texture.get_w(), texture.get_h()))
        )

    def blit_i(self, texture: Texture, src_rect: any = None, dst_rect: any = None) -> None:
        SDL_RenderCopy(
            self.renderer,
            texture.texture,
            src_rect and SDL_Rect(int(src_rect[0]), int(src_rect[1]), int(src_rect[2]), int(src_rect[3])),
            dst_rect and (SDL_Rect(int(dst_rect[0]), int(dst_rect[1]), int(dst_rect[2]), int(dst_rect[3]))
                          if len(dst_rect) > 2 else
                          SDL_Rect(int(dst_rect[0]), int(dst_rect[1]), texture.get_w(), texture.get_h()))
        )

    def blit_ex(
            self, texture: Texture, src_rect: any = None, dst_rect: any = None, angle: float = 0.0,
            center: any = None, flip_horizontal: bool = False, flip_vertical: bool = False
    ) -> None:
        SDL_RenderCopyExF(
            self.renderer,
            texture.texture,
            src_rect and SDL_FRect(src_rect[0], src_rect[1], src_rect[2], src_rect[3]),
            dst_rect and (SDL_FRect(dst_rect[0], dst_rect[1], dst_rect[2], dst_rect[3]) if len(dst_rect) > 2 else
                          SDL_FRect(dst_rect[0], dst_rect[1], texture.get_w(), texture.get_h())),
            angle,
            center and SDL_FPoint(center[0], center[1]),
            ((flip_horizontal and SDL_FLIP_HORIZONTAL) | (flip_vertical and SDL_FLIP_VERTICAL)) or SDL_FLIP_NONE
        )

    def blit_ex_i(
            self, texture: Texture, src_rect: any = None, dst_rect: any = None, angle: float = 0.0,
            center: any = None, flip_horizontal: bool = False, flip_vertical: bool = False
    ) -> None:
        SDL_RenderCopyEx(
            self.renderer,
            texture.texture,
            src_rect and SDL_Rect(int(src_rect[0]), int(src_rect[1]), int(src_rect[2]), int(src_rect[3])),
            dst_rect and (SDL_Rect(int(dst_rect[0]), int(dst_rect[1]), int(dst_rect[2]), int(dst_rect[3]))
                          if len(dst_rect) > 2 else
                          SDL_Rect(int(dst_rect[0]), int(dst_rect[1]), texture.get_w(), texture.get_h())),
            angle,
            center and SDL_FPoint(center[0], center[1]),
            ((flip_horizontal and SDL_FLIP_HORIZONTAL) | (flip_vertical and SDL_FLIP_VERTICAL)) or SDL_FLIP_NONE
        )

    def texture_from_surface(self, surf: Surface) -> Texture:
        return Texture(SDL_CreateTextureFromSurface(self.renderer, surf.surface), self)

    def draw_bezier(self, color: any, points: any, s: float) -> None:
        bezierRGBA(
            self.renderer,
            (ctypes.c_short * len(points))(*(int(point[0]) for point in points)),
            (ctypes.c_short * len(points))(*(int(point[1]) for point in points)),
            len(points), int(s), int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def draw_polygon(self, color: any, points: any, aa: bool = False) -> None:
        (aapolygonRGBA if aa else polygonRGBA)(
            self.renderer,
            (ctypes.c_short * len(points))(*(int(point[0]) for point in points)),
            (ctypes.c_short * len(points))(*(int(point[1]) for point in points)),
            len(points), int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def draw_textured_polygon(self, points: any, surf: Surface, offset: any = (0, 0)) -> None:
        texturedPolygon(
            self.renderer,
            (ctypes.c_short * len(points))(*(int(point[0]) for point in points)),
            (ctypes.c_short * len(points))(*(int(point[1]) for point in points)),
            len(points), surf.surface,
            int(offset[0]), int(offset[1])
        )

    def fill_polygon(self, color: any, points: any) -> None:
        filledPolygonRGBA(
            self.renderer,
            (ctypes.c_short * len(points))(*(int(point[0]) for point in points)),
            (ctypes.c_short * len(points))(*(int(point[1]) for point in points)),
            len(points), int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def draw_trigon(self, color: any, pos1: any, pos2: any, pos3: any, aa: bool = False) -> None:
        (aatrigonRGBA if aa else trigonRGBA)(
            self.renderer,
            int(pos1[0]), int(pos1[1]), int(pos2[0]), int(pos2[1]), int(pos3[0]), int(pos3[1]),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def fill_trigon(self, color: any, pos1: any, pos2: any, pos3: any) -> None:
        filledTrigonRGBA(
            self.renderer,
            int(pos1[0]), int(pos1[1]), int(pos2[0]), int(pos2[1]), int(pos3[0]), int(pos3[1]),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def fill_pie(self, color: any, center: any, rad: float, start: float, end: float) -> None:
        filledPieRGBA(
            self.renderer,
            int(center[0]), int(center[1]), int(rad), int(start), int(end),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def fill_pie_tl(self, color: any, pos: any, rad: float, start: float, end: float) -> None:
        filledPieRGBA(
            self.renderer,
            int(pos[0] + rad), int(pos[1] + rad), int(rad), int(start), int(end),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def draw_pie(self, color: any, center: any, rad: float, start: float, end: float) -> None:
        pieRGBA(
            self.renderer,
            int(center[0]), int(center[1]), int(rad), int(start), int(end),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def draw_pie_tl(self, color: any, pos: any, rad: float, start: float, end: float) -> None:
        pieRGBA(
            self.renderer,
            int(pos[0] + rad), int(pos[1] + rad), int(rad), int(start), int(end),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def draw_arc(self, color: any, center: any, rad: float, start: float, end: float) -> None:
        arcRGBA(
            self.renderer,
            int(center[0]), int(center[1]), int(rad), int(start), int(end),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def draw_arc_tl(self, color: any, pos: any, rad: float, start: float, end: float) -> None:
        arcRGBA(
            self.renderer,
            int(pos[0] + rad), int(pos[1] + rad), int(rad), int(start), int(end),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def fill_circle(self, color: any, center: any, r: float) -> None:
        filledCircleRGBA(
            self.renderer,
            int(center[0]), int(center[1]), int(r),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def fill_circle_tl(self, color: any, pos: any, r: float) -> None:
        filledCircleRGBA(
            self.renderer,
            int(pos[0] + r), int(pos[1] + r), int(r),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def draw_circle_tl(self, color: any, pos: any, r: float, aa: bool = False) -> None:
        (aacircleRGBA if aa else circleRGBA)(
            self.renderer,
            int(pos[0] + r), int(pos[1] + r), int(r),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def draw_circle(self, color: any, center: any, r: float, aa: bool = False) -> None:
        (aacircleRGBA if aa else circleRGBA)(
            self.renderer,
            int(center[0]), int(center[1]), int(r),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def draw_ellipse(self, color: any, center: any, rx: float, ry: float, aa: bool = False) -> None:
        (aaellipseRGBA if aa else ellipseRGBA)(
            self.renderer,
            int(center[0]), int(center[1]), int(rx), int(ry),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def draw_ellipse_tl(self, color: any, pos: any, rx: float, ry: float, aa: bool = False) -> None:
        (aaellipseRGBA if aa else ellipseRGBA)(
            self.renderer,
            int(pos[0] + rx), int(pos[1] + ry), int(rx), int(ry),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def fill_ellipse(self, color: any, center: any, rx: float, ry: float) -> None:
        filledEllipseRGBA(
            self.renderer,
            int(center[0]), int(center[1]), int(rx), int(ry),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def fill_ellipse_tl(self, color: any, pos: any, rx: float, ry: float) -> None:
        filledEllipseRGBA(
            self.renderer,
            int(pos[0] + rx), int(pos[1] + ry), int(rx), int(ry),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def draw_line(self, color: any, start: any, end: any, width: float = 1.0) -> None:
        if width >= 2:
            return self.gfx_draw_thick_line(color, start, end, width)
        return self.sdl_draw_line(color, start, end)

    def draw_point(self, color: any, point: any) -> None:
        self.set_draw_color(color)
        SDL_RenderDrawPointF(self.renderer, point[0], point[1])

    def draw_point_i(self, color: any, point: any) -> None:
        self.set_draw_color(color)
        SDL_RenderDrawPoint(self.renderer, int(point[0]), int(point[1]))

    def draw_points(self, color: any, points: any) -> None:
        self.set_draw_color(color)
        SDL_RenderDrawPointsF(
            self.renderer,
            (SDL_FPoint * len(points))(*(SDL_FPoint(point[0], point[1]) for point in points)),
            len(points)
        )

    def draw_points_i(self, color: any, points: any) -> None:
        self.set_draw_color(color)
        SDL_RenderDrawPoints(
            self.renderer,
            (SDL_Point * len(points))(*(SDL_Point(int(point[0]), int(point[1])) for point in points)),
            len(points)
        )

    def draw_rect(self, color: any, draw_rect: any, rad: float = 0.0) -> None:
        if rad >= 1:
            return self.gfx_draw_rounded_rect(color, draw_rect, rad)
        return self.sdl_draw_rect(color, draw_rect)

    def fill_rect(self, color: any, fill_rect: any, rad: float = 0.0) -> None:
        if rad >= 1:
            return self.gfx_fill_rounded_rect(color, fill_rect, rad)
        return self.sdl_fill_rect(color, fill_rect)

    def gfx_draw_rounded_rect(self, color: any, draw_rect: any, rad: float) -> None:
        roundedRectangleRGBA(
            self.renderer,
            int(draw_rect[0]), int(draw_rect[1]), int(draw_rect[0] + draw_rect[2]), int(draw_rect[1] + draw_rect[3]),
            int(rad), int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def gfx_fill_rounded_rect(self, color: any, fill_rect: any, rad: float) -> None:
        roundedBoxRGBA(
            self.renderer,
            int(fill_rect[0]), int(fill_rect[1]), int(fill_rect[0] + fill_rect[2]), int(fill_rect[1] + fill_rect[3]),
            int(rad), int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def sdl_draw_rect(self, color: any, draw_rect: any = None) -> None:
        self.set_draw_color(color)
        SDL_RenderDrawRectF(
            self.renderer,
            draw_rect and SDL_FRect(draw_rect[0], draw_rect[1], draw_rect[2], draw_rect[3])
        )

    def sdl_draw_rect_i(self, color: any, draw_rect: any = None) -> None:
        self.set_draw_color(color)
        SDL_RenderDrawRect(
            self.renderer,
            draw_rect and SDL_Rect(int(draw_rect[0]), int(draw_rect[1]), int(draw_rect[2]), int(draw_rect[3]))
        )

    def sdl_draw_rects(self, color: any, draw_rects: any) -> None:
        self.set_draw_color(color)
        SDL_RenderDrawRectsF(
            self.renderer,
            (SDL_FRect * len(draw_rects))(*(draw_rect and SDL_FRect(
                draw_rect[0], draw_rect[1], draw_rect[2], draw_rect[3]) for draw_rect in draw_rects)),
            len(draw_rects)
        )

    def sdl_draw_rects_i(self, color: any, draw_rects: any) -> None:
        self.set_draw_color(color)
        SDL_RenderDrawRects(
            self.renderer,
            (SDL_Rect * len(draw_rects))(*(draw_rect and SDL_Rect(int(
                draw_rect[0]), int(draw_rect[1]), int(draw_rect[2]), int(draw_rect[3])) for draw_rect in draw_rects)),
            len(draw_rects)
        )

    def sdl_fill_rect(self, color: any, fill_rect: any = None) -> None:
        self.set_draw_color(color)
        SDL_RenderFillRectF(
            self.renderer,
            fill_rect and SDL_FRect(fill_rect[0], fill_rect[1], fill_rect[2], fill_rect[3])
        )

    def sdl_fill_rect_i(self, color: any, fill_rect: any = None) -> None:
        self.set_draw_color(color)
        SDL_RenderFillRect(
            self.renderer,
            fill_rect and SDL_Rect(int(fill_rect[0]), int(fill_rect[1]), int(fill_rect[2]), int(fill_rect[3]))
        )

    def sdl_fill_rects(self, color: any, fill_rects: any) -> None:
        self.set_draw_color(color)
        SDL_RenderFillRectsF(
            self.renderer,
            (SDL_FRect * len(fill_rects))(*(fill_rect and SDL_FRect(
                fill_rect[0], fill_rect[1], fill_rect[2], fill_rect[3]) for fill_rect in fill_rects)),
            len(fill_rects)
        )

    def sdl_fill_rects_i(self, color: any, fill_rects: any) -> None:
        self.set_draw_color(color)
        SDL_RenderFillRects(
            self.renderer,
            (SDL_Rect * len(fill_rects))(*(fill_rect and SDL_Rect(int(
                fill_rect[0]), int(fill_rect[1]), int(fill_rect[2]), int(fill_rect[3])) for fill_rect in fill_rects)),
            len(fill_rects)
        )

    def gfx_draw_line(self, color: any, start: any, end: any, aa: bool = False) -> None:
        (aalineRGBA if aa else lineRGBA)(
            self.renderer,
            int(start[0]), int(start[1]), int(end[0]), int(end[1]),
            int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def gfx_draw_thick_line(self, color: any, start: any, end: any, width: float) -> None:
        thickLineRGBA(
            self.renderer,
            int(start[0]), int(start[1]), int(end[0]), int(end[1]),
            int(width), int(color[0]), int(color[1]), int(color[2]), int(color[3]) if len(color) > 3 else 255
        )

    def sdl_draw_line(self, color: any, start: any, end: any) -> None:
        self.set_draw_color(color)
        SDL_RenderDrawLineF(self.renderer, start[0], start[1], end[0], end[1])

    def sdl_draw_line_i(self, color: any, start: any, end: any) -> None:
        self.set_draw_color(color)
        SDL_RenderDrawLine(self.renderer, int(start[0]), int(start[1]), int(end[0]), int(end[1]))

    def sdl_draw_lines(self, color: any, points: any) -> None:
        self.set_draw_color(color)
        SDL_RenderDrawLinesF(
            self.renderer,
            (SDL_FPoint * len(points))(*(SDL_FPoint(point[0], point[1]) for point in points)),
            len(points)
        )

    def sdl_draw_lines_i(self, color: any, points: any) -> None:
        self.set_draw_color(color)
        SDL_RenderDrawLines(
            self.renderer,
            (SDL_Point * len(points))(*(SDL_Point(int(point[0]), int(point[1])) for point in points)),
            len(points)
        )

    def clear(self, color: any = (0, 0, 0, 255)) -> None:
        self.set_draw_color(color)
        SDL_RenderClear(self.renderer)

    def set_draw_color(self, color: any) -> None:
        if len(color) > 3:
            SDL_SetRenderDrawColor(self.renderer, int(color[0]), int(color[1]), int(color[2]), int(color[3]))
            SDL_SetRenderDrawBlendMode(self.renderer, SDL_BLENDMODE_NONE if color[3] >= 255 else SDL_BLENDMODE_BLEND)
        else:
            SDL_SetRenderDrawColor(self.renderer, int(color[0]), int(color[1]), int(color[2]), 255)
            SDL_SetRenderDrawBlendMode(self.renderer, SDL_BLENDMODE_NONE)

    def set_blend_mode(self, mode: str) -> None:
        SDL_SetRenderDrawBlendMode(self.renderer, self.app.blend_map[mode])

    def set_vsync(self, vsync: bool) -> None:
        self.vsync = vsync
        SDL_RenderSetVSync(self.renderer, vsync)

    def flip(self) -> None:
        SDL_RenderPresent(self.renderer)

    def get_output_size(self) -> tuple:
        w_ptr, h_ptr = ctypes.c_int(), ctypes.c_int()
        SDL_GetRendererOutputSize(self.renderer, w_ptr, h_ptr)
        return w_ptr.value, h_ptr.value

    def set_as_target(self) -> None:
        self.renderer.set_target(self)

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        SDL_DestroyRenderer(self.renderer)
        self.destroyed = True
        self.blit = None
        self.blit_ex = None
        self.draw_point = None
        self.draw_points = None
        self.sdl_draw_rect = None
        self.sdl_draw_rects = None
        self.sdl_fill_rect = None
        self.sdl_fill_rects = None
        self.sdl_draw_line = None
        self.sdl_draw_lines = None
        del self.window
        del self.app
        return False

    def __del__(self) -> None:
        self.destroy()



import os
import ctypes
from .exceptions import FlagNotFoundError, SDLError
from .events import CommonEvent, QuitEvent, AudioDeviceEvent, DropEvent, TouchFingerEvent, KeyboardEvent,\
    MouseMotionEvent, MouseButtonEvent, MouseWheelEvent, TextEditingEvent, TextInputEvent, DisplayEvent, WindowEvent,\
    JoyAxisEvent, JoyBallEvent, JoyButtonEvent, JoyDeviceEvent, JoyHatEvent, JoyBatteryEvent, ControllerAxisEvent,\
    ControllerButtonEvent, ControllerDeviceEvent, ControllerTouchpadEvent, ControllerSensorEvent
from .sdl import SDLVersion, sdl_dir
from .surface import Surface, SurfaceAnimation
from .video import PixelFormat
from .window import Window
from .touch import TouchDevice
from .joystick import Joystick
from .gamecontroller import GameController
from .sensor import Sensor
from sdl2 import *

try:
    from sdl2.sdlimage import *
except Exception as _err:
    print(f'Failed to import SDL2_image [{_err}]. Image loading will be disabled!')
try:
    from sdl2.sdlmixer import *
except Exception as _err:
    print(f'Failed to import SDL2_mixer [{_err}]. Mixer will be disabled!')
try:
    from sdl2.sdlttf import *
except Exception as _err:
    print(f'Failed to import SDL2_ttf [{_err}]. Font loading and rendering will be disabled!')

try:
    SDL_TEXTEDITING_EXT = SDL_TEXTEDITING_EXT
except NameError:
    SDL_TEXTEDITING_EXT = 0x305
try:
    SDL_DISPLAYEVENT = SDL_DISPLAYEVENT
except NameError:
    SDL_DISPLAYEVENT = 0x150
try:
    SDL_LOCALECHANGED = SDL_LOCALECHANGED
    SDL_KEYMAPCHANGED = SDL_KEYMAPCHANGED
    SDL_CLIPBOARDUPDATE = SDL_CLIPBOARDUPDATE
except NameError:
    SDL_LOCALECHANGED = 0x107
    SDL_KEYMAPCHANGED = 0x304
    SDL_CLIPBOARDUPDATE = 0x900
try:
    SDL_JOYBATTERYUPDATED = SDL_JOYBATTERYUPDATED
except NameError:
    SDL_JOYBATTERYUPDATED = 0x607
try:
    SDL_CONTROLLERTOUCHPADDOWN = SDL_CONTROLLERTOUCHPADDOWN
    SDL_CONTROLLERTOUCHPADMOTION = SDL_CONTROLLERTOUCHPADMOTION
    SDL_CONTROLLERTOUCHPADUP = SDL_CONTROLLERTOUCHPADUP
    SDL_CONTROLLERSENSORUPDATE = SDL_CONTROLLERSENSORUPDATE
except NameError:
    SDL_CONTROLLERTOUCHPADDOWN = 0x656
    SDL_CONTROLLERTOUCHPADMOTION = 0x657
    SDL_CONTROLLERTOUCHPADUP = 0x658
    SDL_CONTROLLERSENSORUPDATE = 0x659
try:
    SDL_INIT_SENSOR = SDL_INIT_SENSOR
except NameError:
    SDL_INIT_SENSOR = 0x00008000
try:
    IMG_INIT_JXL = IMG_INIT_JXL
    IMG_INIT_AVIF = IMG_INIT_AVIF
except NameError:
    IMG_INIT_JXL = 0x00000010
    IMG_INIT_AVIF = 0x00000020
try:
    MIX_INIT_MID = MIX_INIT_MID
    MIX_INIT_OPUS = MIX_INIT_OPUS
except NameError:
    MIX_INIT_MID = 0x00000020
    MIX_INIT_OPUS = 0x00000040


class App:
    def __init__(self) -> None:
        self.destroyed = True
        self.encoding = 'utf-8'
        if SDL_BYTEORDER == SDL_LIL_ENDIAN:
            self.endian = 'little'
            self.default_rgb_mask = (0x0000FF, 0x00FF00, 0xFF0000, 0)
            self.default_rgba_mask = (0x000000FF, 0x0000FF00, 0x00FF0000, 0xFF000000)
        else:
            self.endian = 'big'
            self.default_rgb_mask = (0xFF0000, 0x00FF00, 0x0000FF, 0)
            self.default_rgba_mask = (0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF)
        self.cwd = os.getcwd()
        self.init_flags = {
            'sdl': 0,
            'image': 0,
            'mixer': 0,
            'has_sdl': False,
            'has_image': False,
            'has_mixer': False,
            'has_ttf': False
        }
        self.blend_map = {
            'none': SDL_BLENDMODE_NONE,
            'blend': SDL_BLENDMODE_BLEND,
            'add': SDL_BLENDMODE_ADD,
            'mod': SDL_BLENDMODE_MOD
        }
        try:
            self.blend_map['mul'] = SDL_BLENDMODE_MUL
        except NameError:
            self.blend_map['mul'] = 0x00000008
        self.r_blend_map = {b: a for a, b in self.blend_map.items()}
        self.event_map = {
            SDL_AUDIODEVICEADDED: lambda: self.on_audio_device_add(AudioDeviceEvent(self.sdl_event.adevice)),
            SDL_AUDIODEVICEREMOVED: lambda: self.on_audio_device_remove(AudioDeviceEvent(self.sdl_event.adevice)),
            SDL_QUIT: lambda: self.on_quit(QuitEvent(self.sdl_event.quit)),
            SDL_DROPFILE: lambda: self.on_drop_file(DropEvent(self.sdl_event.drop, self)),
            SDL_DROPTEXT: lambda: self.on_drop_text(DropEvent(self.sdl_event.drop, self)),
            SDL_DROPBEGIN: lambda: self.on_drop_begin(DropEvent(self.sdl_event.drop, self)),
            SDL_DROPCOMPLETE: lambda: self.on_drop_complete(DropEvent(self.sdl_event.drop, self)),
            SDL_FINGERMOTION: lambda: self.on_finger_move(TouchFingerEvent(self.sdl_event.tfinger, self)),
            SDL_FINGERDOWN: lambda: self.on_finger_down(TouchFingerEvent(self.sdl_event.tfinger, self)),
            SDL_FINGERUP: lambda: self.on_finger_up(TouchFingerEvent(self.sdl_event.tfinger, self)),
            SDL_KEYDOWN: lambda: self.on_key_down(KeyboardEvent(self.sdl_event.key, self)),
            SDL_KEYUP: lambda: self.on_key_up(KeyboardEvent(self.sdl_event.key, self)),
            SDL_TEXTEDITING: lambda: self.on_text_edit(TextEditingEvent(self.sdl_event.edit, self)),
            SDL_MOUSEMOTION: lambda: self.on_mouse_move(MouseMotionEvent(self.sdl_event.motion, self)),
            SDL_MOUSEBUTTONDOWN: lambda: self.on_mouse_down(MouseButtonEvent(self.sdl_event.button, self)),
            SDL_MOUSEBUTTONUP: lambda: self.on_mouse_up(MouseButtonEvent(self.sdl_event.button, self)),
            SDL_MOUSEWHEEL: lambda: self.on_mouse_wheel(MouseWheelEvent(self.sdl_event.wheel, self)),
            SDL_TEXTINPUT: lambda: self.on_text_input(TextInputEvent(self.sdl_event.text, self)),
            SDL_WINDOWEVENT: lambda: self.on_window_event(WindowEvent(self.sdl_event.window, self)),
            SDL_APP_TERMINATING: lambda: self.on_terminate(CommonEvent(self.sdl_event.common)),
            SDL_APP_LOWMEMORY: lambda: self.on_low_memory(CommonEvent(self.sdl_event.common)),
            SDL_APP_WILLENTERBACKGROUND: lambda: self.on_will_enter_background(CommonEvent(self.sdl_event.common)),
            SDL_APP_DIDENTERBACKGROUND: lambda: self.on_did_enter_background(CommonEvent(self.sdl_event.common)),
            SDL_APP_WILLENTERFOREGROUND: lambda: self.on_will_enter_foreground(CommonEvent(self.sdl_event.common)),
            SDL_APP_DIDENTERFOREGROUND: lambda: self.on_did_enter_foreground(CommonEvent(self.sdl_event.common)),
            SDL_RENDER_TARGETS_RESET: lambda: self.on_render_targets_reset(CommonEvent(self.sdl_event.common)),
            SDL_RENDER_DEVICE_RESET: lambda: self.on_render_device_reset(CommonEvent(self.sdl_event.common)),
            SDL_TEXTEDITING_EXT: lambda: self.on_text_edit_ext(TextEditingEvent(self.sdl_event.edit, self)),
            SDL_DISPLAYEVENT: lambda: self.on_display_event(DisplayEvent(self.sdl_event.display, self)),
            SDL_LOCALECHANGED: lambda: self.on_locale_change(CommonEvent(self.sdl_event.common)),
            SDL_KEYMAPCHANGED: lambda: self.on_keymap_change(CommonEvent(self.sdl_event.common)),
            SDL_CLIPBOARDUPDATE: lambda: self.on_clipboard_update(CommonEvent(self.sdl_event.common)),
            SDL_JOYAXISMOTION: lambda: self.on_joy_axis_move(JoyAxisEvent(self.sdl_event.jaxis)),
            SDL_JOYBALLMOTION: lambda: self.on_joy_ball_move(JoyBallEvent(self.sdl_event.jball)),
            SDL_JOYBUTTONDOWN: lambda: self.on_joy_button_down(JoyButtonEvent(self.sdl_event.jbutton)),
            SDL_JOYBUTTONUP: lambda: self.on_joy_button_up(JoyButtonEvent(self.sdl_event.jbutton)),
            SDL_JOYDEVICEADDED: lambda: self.on_joy_device_add(JoyDeviceEvent(self.sdl_event.jdevice)),
            SDL_JOYDEVICEREMOVED: lambda: self.on_joy_device_remove(JoyDeviceEvent(self.sdl_event.jdevice)),
            SDL_JOYHATMOTION: lambda: self.on_joy_hat_move(JoyHatEvent(self.sdl_event.jhat)),
            SDL_JOYBATTERYUPDATED: lambda: self.on_joy_battery_update(JoyBatteryEvent(self.sdl_event.jbattery)),
            SDL_CONTROLLERAXISMOTION: lambda: self.on_controller_axis_move(ControllerAxisEvent(self.sdl_event.caxis)),
            SDL_CONTROLLERBUTTONDOWN: lambda: self.on_controller_button_down(
                ControllerButtonEvent(self.sdl_event.cbutton)
            ),
            SDL_CONTROLLERBUTTONUP: lambda: self.on_controller_button_up(
                ControllerButtonEvent(self.sdl_event.cbutton)
            ),
            SDL_CONTROLLERDEVICEADDED: lambda: self.on_controller_device_add(
                ControllerDeviceEvent(self.sdl_event.cdevice)
            ),
            SDL_CONTROLLERDEVICEREMOVED: lambda: self.on_controller_device_remove(
                ControllerDeviceEvent(self.sdl_event.cdevice)
            ),
            SDL_CONTROLLERDEVICEREMAPPED: lambda: self.on_controller_device_remap(
                ControllerDeviceEvent(self.sdl_event.cdevice)
            ),
            SDL_CONTROLLERTOUCHPADDOWN: lambda: self.on_controller_touchpad_down(
                ControllerTouchpadEvent(self.sdl_event.cdevice)
            ),
            SDL_CONTROLLERTOUCHPADMOTION: lambda: self.on_controller_touchpad_move(
                ControllerTouchpadEvent(self.sdl_event.cdevice)
            ),
            SDL_CONTROLLERTOUCHPADUP: lambda: self.on_controller_touchpad_up(
                ControllerTouchpadEvent(self.sdl_event.cdevice)
            ),
            SDL_CONTROLLERSENSORUPDATE: lambda: self.on_controller_sensor_update(
                ControllerSensorEvent(self.sdl_event.cdevice)
            )
        }
        self.windows = {}
        self.running = False
        self.rel_mouse_mode = False
        self.mouse_capture = False
        self.platform = self.bts(SDL_GetPlatform())
        self.keyboard_state_ptr = SDL_GetKeyboardState(None)
        self.sdl_event = SDL_Event()
        self.destroyed = False
        # TODO:
        #  RW ops (do we need them?)
        #  hidapi (do we need it?), haptics, gestures (do we need them?)
        #  pixels (palettes, etc)
        #  Custom message box

    def get_sensors(self) -> tuple:
        return tuple(Sensor(self, _x) for _x in range(SDL_NumSensors()))

    def get_joysticks(self) -> tuple:
        return tuple(Joystick(self, _x) for _x in range(SDL_NumJoysticks()))

    def get_game_controllers(self) -> tuple:
        return tuple(GameController(self, _x) for _x in range(SDL_NumJoysticks()) if SDL_IsGameController(_x))

    def get_touch_devices(self) -> tuple:
        return tuple(TouchDevice(self, _x) for _x in range(SDL_GetNumTouchDevices()))

    def get_key_state(self, key: str) -> bool:
        return bool(self.keyboard_state_ptr[SDL_GetScancodeFromName(self.stb(key))])

    @staticmethod
    def get_img_version() -> tuple:
        ver = IMG_Linked_Version().contents
        return ver.major, ver.minor, ver.patch

    def set_mouse_capture(self, enabled: bool) -> None:
        self.mouse_capture = enabled
        SDL_CaptureMouse(enabled)

    def set_rel_mouse(self, enabled: bool) -> None:
        self.rel_mouse_mode = enabled
        SDL_SetRelativeMouseMode(enabled)

    @staticmethod
    def set_mouse_pos(pos: any) -> None:
        SDL_WarpMouseGlobal(int(pos[0]), int(pos[1]))

    @staticmethod
    def get_mouse_buttons() -> tuple:
        state = SDL_GetGlobalMouseState(None, None)
        return (
            bool(state & SDL_BUTTON_LMASK),
            bool(state & SDL_BUTTON_MMASK),
            bool(state & SDL_BUTTON_RMASK),
            bool(state & SDL_BUTTON_X1MASK),
            bool(state & SDL_BUTTON_X2MASK)
        )

    @staticmethod
    def get_mouse_pos() -> tuple:
        x_ptr, y_ptr = ctypes.c_int(), ctypes.c_int()
        SDL_GetGlobalMouseState(x_ptr, y_ptr)
        return x_ptr.value, y_ptr.value

    @staticmethod
    def get_rel_mouse_buttons() -> tuple:
        state = SDL_GetRelativeMouseState(None, None)
        return (
            bool(state & SDL_BUTTON_LMASK),
            bool(state & SDL_BUTTON_MMASK),
            bool(state & SDL_BUTTON_RMASK),
            bool(state & SDL_BUTTON_X1MASK),
            bool(state & SDL_BUTTON_X2MASK)
        )

    @staticmethod
    def get_rel_mouse_pos() -> tuple:
        x_ptr, y_ptr = ctypes.c_int(), ctypes.c_int()
        SDL_GetRelativeMouseState(x_ptr, y_ptr)
        return x_ptr.value, y_ptr.value

    @staticmethod
    def get_power_info() -> dict:
        seconds_ptr, percent_ptr = ctypes.c_int(), ctypes.c_int()
        power_state = SDL_GetPowerInfo(seconds_ptr, percent_ptr)
        return {
            'power_state': {
                SDL_POWERSTATE_UNKNOWN: 'unknown',
                SDL_POWERSTATE_ON_BATTERY: 'on_battery',
                SDL_POWERSTATE_NO_BATTERY: 'no_battery',
                SDL_POWERSTATE_CHARGING: 'charging',
                SDL_POWERSTATE_CHARGED: 'charged'
            }.get(power_state),
            'seconds_left': None if seconds_ptr.value == -1 else seconds_ptr.value,
            'percentage': None if percent_ptr.value == -1 else percent_ptr.value
        }

    def get_pref_path(self, app: str, org: str = None) -> str:
        pref_path = SDL_GetPrefPath(org and self.stb(org), app and self.stb(app))
        if not pref_path:
            self.raise_error()
        return self.bts(pref_path)

    def get_base_path(self) -> str:
        base_path = SDL_GetBasePath()
        if not base_path:
            self.raise_error()
        return self.bts(base_path)

    @staticmethod
    def get_keyboard_state() -> tuple:
        return bool(SDL_IsTextInputActive()), bool(SDL_IsTextInputShown()), bool(SDL_HasScreenKeyboardSupport())

    @staticmethod
    def clear_text_input() -> None:
        SDL_ClearComposition()

    @staticmethod
    def enable_text_input(enabled: bool) -> None:
        (SDL_StartTextInput if enabled else SDL_StopTextInput)()

    def set_clipboard_text(self, text: str) -> None:
        SDL_SetClipboardText(self.stb(text))

    def set_primary_selection_text(self, text: str) -> None:
        SDL_SetPrimarySelectionText(self.stb(text))

    def get_clipboard_text(self) -> str:
        if not SDL_HasClipboardText():
            return ''
        return self.bts(SDL_GetClipboardText())

    def get_primary_selection_text(self) -> str:
        if not SDL_HasPrimarySelectionText():
            return ''
        return self.bts(SDL_GetPrimarySelectionText())

    def open_url(self, url: str) -> None:
        if SDL_OpenURL(self.stb(url)) < 0:
            self.raise_error()

    def show_message_box(
            self,
            title: str,
            text: str,
            icon_type: str = None,
            window: Window = None
    ) -> None:
        flags = 0
        if icon_type == 'information':
            flags |= SDL_MESSAGEBOX_INFORMATION
        elif icon_type == 'warning':
            flags |= SDL_MESSAGEBOX_WARNING
        elif icon_type == 'error':
            flags |= SDL_MESSAGEBOX_ERROR
        if SDL_ShowSimpleMessageBox(flags, self.stb(title), self.stb(text), window and window.window) < 0:
            self.raise_error()

    def get_preferred_locales(self) -> tuple:
        return tuple(
            {'lang': self.bts(_x.language), 'country': self.bts(_x.country)} for _x in SDL_GetPreferredLocales()
        )

    def get_sdl_info(self) -> dict:
        ver_ptr = SDL_version()
        SDL_GetVersion(ver_ptr)
        return {
            'version': (ver_ptr.major, ver_ptr.minor, ver_ptr.patch),
            'revision': self.bts(SDL_GetRevision())
        }

    @staticmethod
    def get_cpu_info() -> dict:
        return {
            'count': SDL_GetCPUCount(),
            'cache_line_size': SDL_GetCPUCacheLineSize(),
            'ram': SDL_GetSystemRAM(),
            'simd_alignment': SDL_SIMDGetAlignment(),
            'RDTSC': bool(SDL_HasRDTSC()),
            'AltiVec': bool(SDL_HasAltiVec()),
            'MMX': bool(SDL_HasMMX()),
            '3DNow': bool(SDL_Has3DNow()),
            'SSE': bool(SDL_HasSSE()),
            'SSE2': bool(SDL_HasSSE2()),
            'SSE3': bool(SDL_HasSSE3()),
            'SSE41': bool(SDL_HasSSE41()),
            'SSE42': bool(SDL_HasSSE42()),
            'AVX': bool(SDL_HasAVX()),
            'AVX2': bool(SDL_HasAVX2()),
            'AVX512F': bool(SDL_HasAVX512F()),
            'ARMSIMD': bool(SDL_HasARMSIMD()),
            'NEON': bool(SDL_HasNEON()),
            'LSX': bool(SDL_HasLSX()),
            'LASX': bool(SDL_HasLASX())
        }

    @staticmethod
    def alloc_pixel_format(pixel_format: PixelFormat) -> SDL_PixelFormat:
        return SDL_AllocFormat(pixel_format.pixel_format)

    def surface_from_bytes(
            self, data: any, size: any, depth: int, pitch: int,
            mask: any = (0, 0, 0, 0), pixel_format: PixelFormat = None
    ) -> Surface:
        if pixel_format:
            surf = SDL_CreateRGBSurfaceWithFormatFrom(
                data, int(size[0]), int(size[1]), depth, pitch, mask[0],
                mask[1], mask[2], mask[3], pixel_format.pixel_format
            )
        else:
            surf = SDL_CreateRGBSurfaceFrom(
                data, int(size[0]), int(size[1]), depth, pitch, mask[0], mask[1], mask[2], mask[3]
            )
        if not surf:
            self.raise_error()
        return Surface(surf, self)

    def create_rgb_surface(
            self, size: any, depth: int = 32, mask: any = (0, 0, 0, 0), pixel_format: PixelFormat = None
    ) -> Surface:
        if pixel_format:
            surf = SDL_CreateRGBSurfaceWithFormat(
                0, int(size[0]), int(size[1]), depth, mask[0], mask[1], mask[2], mask[3], pixel_format.pixel_format
            )
        else:
            surf = SDL_CreateRGBSurface(0, int(size[0]), int(size[1]), depth, mask[0], mask[1], mask[2], mask[3])
        if not surf:
            self.raise_error()
        return Surface(surf, self)

    def surface_from_bmp(self, path: str) -> Surface:
        surf = SDL_LoadBMP(self.stb(path))
        if not surf:
            self.raise_error()
        return Surface(surf, self)

    def surface_from_file(self, path: str) -> Surface:
        surf = IMG_Load(self.stb(path))
        if not surf:
            self.raise_error(IMG_GetError)
        return Surface(surf, self)

    def animation_from_file(self, path: str) -> SurfaceAnimation:
        anim = IMG_LoadAnimation(self.stb(path))
        if not anim:
            self.raise_error(IMG_GetError)
        return SurfaceAnimation(anim, self)

    def poll_events(self) -> None:
        while SDL_PollEvent(self.sdl_event):
            (self.event_map.get(self.sdl_event.type) or self.on_unknown_event)()

    def on_unknown_event(self) -> None:
        event_names = []
        for var_name in sdl_dir:
            if not var_name.startswith('SDL_'):
                continue
            if eval(var_name) == self.sdl_event.type:
                event_names.append(var_name)
        print(f'Unknown event {self.sdl_event.type}: {", ".join(event_names)}')

    def run_loop(self) -> None:
        self.running = True
        while self.running:
            self.on_tick()

    def stop_loop(self) -> None:
        self.running = False

    def init(
            self,
            sdl_flags_list: any = ('video', 'events', 'timer'),
            image_formats: any = ('png', 'jpg'),
            mixer_formats: any = None,
            init_ttf: bool = True
    ) -> None:
        if sdl_flags_list:
            flag_map = {
                'timer': SDL_INIT_TIMER,
                'audio': SDL_INIT_AUDIO,
                'video': SDL_INIT_VIDEO,
                'joystick': SDL_INIT_JOYSTICK,
                'haptic': SDL_INIT_HAPTIC,
                'game_controller': SDL_INIT_GAMECONTROLLER,
                'events': SDL_INIT_EVENTS,
                'sensor': SDL_INIT_SENSOR
            }
            flags = 0
            for flag_str in sdl_flags_list:
                flag = flag_map.get(flag_str)
                if not flag:
                    raise FlagNotFoundError(f'Flag {flag_str} not found among: {", ".join(flag_map.keys())}')
                flags |= flag
            flags |= self.init_flags['sdl']
            diff = self.init_flags['sdl'] ^ flags
            if diff and SDL_Init(diff) < 0:
                self.raise_error()
            self.init_flags['sdl'] = flags
            self.init_flags['has_sdl'] = True
        if image_formats:
            flag_map = {
                'png': IMG_INIT_PNG,
                'jpg': IMG_INIT_JPG,
                'jxl': IMG_INIT_JXL,
                'tif': IMG_INIT_TIF,
                'avif': IMG_INIT_AVIF,
                'webp': IMG_INIT_WEBP
            }
            flags = 0
            for flag_str in image_formats:
                flag = flag_map.get(flag_str)
                if not flag:
                    raise FlagNotFoundError(f'Format {flag_str} not found among: {", ".join(flag_map.keys())}')
                flags |= flag
            flags |= self.init_flags['image']
            diff = self.init_flags['image'] ^ flags
            if not IMG_Init(diff) == diff:
                self.raise_error(IMG_GetError)
            self.init_flags['image'] = flags
            self.init_flags['has_image'] = True
        if mixer_formats:
            flag_map = {
                'mp3': MIX_INIT_MP3,
                'mid': MIX_INIT_MID,
                'mod': MIX_INIT_MOD,
                'ogg': MIX_INIT_OGG,
                'flac': MIX_INIT_FLAC,
                'opus': MIX_INIT_OPUS
            }
            flags = 0
            for flag_str in mixer_formats:
                flag = flag_map.get(flag_str)
                if not flag:
                    raise FlagNotFoundError(f'Format {flag_str} not found among: {", ".join(flag_map.keys())}')
                flags |= flag
            flags |= self.init_flags['mixer']
            diff = self.init_flags['mixer'] ^ flags
            if not Mix_Init(diff) == diff:
                self.raise_error(Mix_GetError)
            self.init_flags['mixer'] = flags
            self.init_flags['has_mixer'] = True
        if init_ttf and not self.init_flags['has_ttf']:
            if TTF_Init() < 0:
                self.raise_error(TTF_GetError)
            self.init_flags['has_ttf'] = True

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        self.destroyed = False
        if self.init_flags['has_ttf']:
            self.init_flags['has_ttf'] = False
            TTF_Quit()
        if self.init_flags['has_mixer']:
            self.init_flags['has_mixer'] = False
            self.init_flags['mixer'] = 0
            Mix_Quit()
        if self.init_flags['has_image']:
            self.init_flags['has_image'] = False
            self.init_flags['image'] = 0
            IMG_Quit()
        if self.init_flags['has_sdl']:
            self.init_flags['has_sdl'] = False
            self.init_flags['sdl'] = 0
            SDL_Quit()
        return False

    def get_error(self, error_func: any = SDL_GetError) -> str:
        return self.bts(error_func())

    def raise_error(self, error_func: any = SDL_GetError) -> None:
        raise SDLError(self.get_error(error_func))

    def stb(self, str_to_convert: str, encoding: str = None) -> bytes:
        return str_to_convert.encode(encoding or self.encoding, errors='replace')

    def bts(self, bytes_to_convert: bytes, encoding: str = None) -> str:
        return bytes_to_convert.decode(encoding or self.encoding, errors='replace')

    def p(self, *path: str) -> str:
        return os.path.join(self.cwd, *path)

    @staticmethod
    def get_sdl_version() -> SDLVersion:
        ver = SDL_version()
        SDL_GetVersion(ver)
        return SDLVersion(ver)

    @staticmethod
    def print_sub_content(obj_for_displaying: any) -> None:
        for elem_name in dir(obj_for_displaying):
            if elem_name.startswith('_'):
                continue
            elem = getattr(obj_for_displaying, elem_name)
            if not hasattr(elem, '__str__'):
                continue
            print(f'{elem_name}: {elem}')

    def on_tick(self) -> None:
        pass

    def on_audio_device_add(self, event: AudioDeviceEvent) -> None:
        pass

    def on_audio_device_remove(self, event: AudioDeviceEvent) -> None:
        pass

    def on_quit(self, event: QuitEvent) -> None:
        pass

    def on_drop_file(self, event: DropEvent) -> None:  # noqa
        event.window.on_drop_file(event)

    def on_drop_text(self, event: DropEvent) -> None:  # noqa
        event.window.on_drop_text(event)

    def on_drop_begin(self, event: DropEvent) -> None:  # noqa
        event.window.on_drop_begin(event)

    def on_drop_complete(self, event: DropEvent) -> None:  # noqa
        event.window.on_drop_complete(event)

    def on_finger_move(self, event: TouchFingerEvent) -> None:  # noqa
        event.window.on_finger_move(event)

    def on_finger_down(self, event: TouchFingerEvent) -> None:  # noqa
        event.window.on_finger_down(event)

    def on_finger_up(self, event: TouchFingerEvent) -> None:  # noqa
        event.window.on_finger_up(event)

    def on_key_down(self, event: KeyboardEvent) -> None:  # noqa
        event.window.on_key_down(event)

    def on_key_up(self, event: KeyboardEvent) -> None:  # noqa
        event.window.on_key_up(event)

    def on_text_edit(self, event: TextEditingEvent) -> None:  # noqa
        event.window.on_text_edit(event)

    def on_text_edit_ext(self, event: TextEditingEvent) -> None:  # noqa
        event.window.on_text_edit_ext(event)

    def on_mouse_move(self, event: MouseMotionEvent) -> None:  # noqa
        if event.emulated and not event.window.emulate_mouse_with_touch:
            return
        event.window.on_mouse_move(event)

    def on_mouse_down(self, event: MouseButtonEvent) -> None:  # noqa
        if event.emulated and not event.window.emulate_mouse_with_touch:
            return
        event.window.on_mouse_down(event)

    def on_mouse_up(self, event: MouseButtonEvent) -> None:  # noqa
        if event.emulated and not event.window.emulate_mouse_with_touch:
            return
        event.window.on_mouse_up(event)

    def on_mouse_wheel(self, event: MouseWheelEvent) -> None:  # noqa
        if event.emulated and not event.window.emulate_mouse_with_touch:
            return
        event.window.on_mouse_wheel(event)

    def on_text_input(self, event: TextInputEvent) -> None:  # noqa
        event.window.on_text_input(event)

    def on_display_event(self, event: DisplayEvent) -> None:
        pass

    def on_window_event(self, event: WindowEvent) -> None:
        event.window.event_map.get(self.sdl_event.window.event)(event)

    def on_terminate(self, event: CommonEvent) -> None:
        pass

    def on_low_memory(self, event: CommonEvent) -> None:
        pass

    def on_will_enter_background(self, event: CommonEvent) -> None:
        pass

    def on_did_enter_background(self, event: CommonEvent) -> None:
        pass

    def on_will_enter_foreground(self, event: CommonEvent) -> None:
        pass

    def on_did_enter_foreground(self, event: CommonEvent) -> None:
        pass

    def on_locale_change(self, event: CommonEvent) -> None:
        pass

    def on_keymap_change(self, event: CommonEvent) -> None:
        pass

    def on_clipboard_update(self, event: CommonEvent) -> None:
        pass

    def on_render_targets_reset(self, event: CommonEvent) -> None:
        pass

    def on_render_device_reset(self, event: CommonEvent) -> None:
        pass

    def on_joy_axis_move(self, event: JoyAxisEvent) -> None:
        pass

    def on_joy_ball_move(self, event: JoyBallEvent) -> None:
        pass

    def on_joy_button_down(self, event: JoyButtonEvent) -> None:
        pass

    def on_joy_button_up(self, event: JoyButtonEvent) -> None:
        pass

    def on_joy_device_add(self, event: JoyDeviceEvent) -> None:
        pass

    def on_joy_device_remove(self, event: JoyDeviceEvent) -> None:
        pass

    def on_joy_hat_move(self, event: JoyHatEvent) -> None:
        pass

    def on_joy_battery_update(self, event: JoyBatteryEvent) -> None:
        pass

    def on_controller_axis_move(self, event: ControllerAxisEvent) -> None:
        pass

    def on_controller_button_down(self, event: ControllerButtonEvent) -> None:
        pass

    def on_controller_button_up(self, event: ControllerButtonEvent) -> None:
        pass

    def on_controller_device_add(self, event: ControllerDeviceEvent) -> None:
        pass

    def on_controller_device_remove(self, event: ControllerDeviceEvent) -> None:
        pass

    def on_controller_device_remap(self, event: ControllerDeviceEvent) -> None:
        pass

    def on_controller_touchpad_down(self, event: ControllerTouchpadEvent) -> None:
        pass

    def on_controller_touchpad_move(self, event: ControllerTouchpadEvent) -> None:
        pass

    def on_controller_touchpad_up(self, event: ControllerTouchpadEvent) -> None:
        pass

    def on_controller_sensor_update(self, event: ControllerSensorEvent) -> None:
        pass

    def __del__(self) -> None:
        self.destroy()



import ctypes
from sdl2 import *


class Sensor:
    def __init__(self, app: any, index: int) -> None:
        self.destroyed = True
        self.index = index
        self.name = app.bts(SDL_SensorGetDeviceName(index))
        self.type = {
            SDL_SENSOR_INVALID: 'invalid',
            SDL_SENSOR_UNKNOWN: 'unknown',
            SDL_SENSOR_ACCEL: 'accel',
            SDL_SENSOR_GYRO: 'gyro',
            SDL_SENSOR_ACCEL_L: 'accel_l',
            SDL_SENSOR_GYRO_L: 'gyro_l',
            SDL_SENSOR_ACCEL_R: 'accel_r',
            SDL_SENSOR_GYRO_R: 'gyro_r'
        }.get(SDL_SensorGetDeviceType(index))
        self.non_portable_type = SDL_SensorGetDeviceNonPortableType(index)
        self.instance_id = SDL_SensorGetDeviceInstanceID(index)
        self.sensor = SDL_Sensor()
        self.opened = False
        self.destroyed = False

    def get_data(self, num_values: int) -> tuple:
        data_ptr = (ctypes.c_float * num_values)()
        SDL_SensorGetData(self.sensor, data_ptr, num_values)
        return tuple(data_ptr[_x] for _x in range(num_values))

    def open_from_instance_id(self) -> None:
        self.opened = True
        self.sensor = SDL_SensorFromInstanceID(self.instance_id)

    def open(self) -> None:
        self.opened = True
        self.sensor = SDL_SensorOpen(self.index)

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        if self.opened:
            self.opened = False
            SDL_SensorClose(self.sensor)
        self.destroyed = True
        return False

    def __del__(self) -> None:
        self.destroy()



import array
import ctypes
from sdl2 import *

try:
    from sdl2.sdlmixer import *
except:  # noqa
    pass

try:
    MUS_MP3_MAD_UNUSED = MUS_MP3_MAD_UNUSED
except NameError:
    MUS_MP3_MAD_UNUSED = 7
try:
    MUS_MODPLUG_UNUSED = MUS_MODPLUG_UNUSED
except NameError:
    MUS_MODPLUG_UNUSED = 10
try:
    MUS_OPUS = MUS_OPUS
except NameError:
    MUS_OPUS = 11


# TODO:
#  Effects, etc


class Chunk:
    def __init__(self, mixer: any, path: str = None, wav_data: bytes = None, raw_data: bytes = None) -> None:
        self.destroyed = True
        self.app = mixer.app
        self.path = path
        if raw_data:
            self.buffer = array.array('B', raw_data)
            addr, count = self.buffer.buffer_info()
            self.chunk = Mix_QuickLoad_RAW(ctypes.cast(addr, ctypes.POINTER(ctypes.c_ubyte)), len(raw_data))
        elif wav_data:
            self.buffer = array.array('B', wav_data)
            addr, count = self.buffer.buffer_info()
            self.chunk = Mix_QuickLoad_WAV(ctypes.cast(addr, ctypes.POINTER(ctypes.c_ubyte)))
        elif path:
            self.chunk = Mix_LoadWAV(self.app.stb(path))
        else:
            raise RuntimeError('No Chunk Audio Data')
        if not self.chunk:
            self.app.raise_error(Mix_GetError)
        self.channel = -1
        self.destroyed = False

    def is_playing(self) -> bool:
        return bool(Mix_Playing(self.channel))

    def is_paused(self) -> bool:
        return bool(Mix_Paused(self.channel))

    def pause(self) -> None:
        Mix_Pause(self.channel)

    def resume(self) -> None:
        Mix_Resume(self.channel)

    def get_fading_status(self) -> str:
        status = Mix_FadingChannel(self.channel)
        if status == MIX_FADING_OUT:
            return 'out'
        if status == MIX_FADING_IN:
            return 'in'
        return 'no'

    def fade_out(self, fade_out: float) -> None:
        Mix_FadeOutChannel(self.channel, int(fade_out * 1000))

    def expire(self, time: float = -1) -> None:
        Mix_ExpireChannel(self.channel, -1 if time == -1 else int(time * 1000))

    def halt(self) -> None:
        Mix_HaltChannel(self.channel)

    def set_chunk_volume(self, volume: float = 1.0) -> None:
        Mix_VolumeChunk(self.chunk, int(volume * MIX_MAX_VOLUME))

    def set_volume(self, volume: float = 1.0) -> None:
        Mix_Volume(self.channel, int(volume * MIX_MAX_VOLUME))

    def play(self, loops: int = 0, channel: int = -1, fade_in: float = 0.0) -> None:
        if fade_in:
            result = Mix_FadeInChannel(channel, self.chunk, loops, int(fade_in * 1000))
        else:
            result = Mix_PlayChannel(channel, self.chunk, loops)
        if result < 0:
            self.app.raise_error(Mix_GetError)
        self.channel = result

    def play_timed(self, loops: int = 0, time: float = -1, channel: int = -1, fade_in: float = 0.0) -> None:
        if fade_in:
            result = Mix_FadeInChannelTimed(
                channel, self.chunk, loops, -1 if time == -1 else int(time * 1000), int(fade_in * 1000)
            )
        else:
            result = Mix_PlayChannelTimed(channel, self.chunk, loops, -1 if time == -1 else int(time * 1000))
        if result < 0:
            self.app.raise_error(Mix_GetError)
        self.channel = result

    def reverse(self, enabled: bool) -> None:
        Mix_SetReverseStereo(self.channel, enabled)

    def set_position(self, angle: float, distance: float) -> None:
        Mix_SetPosition(self.channel, int(angle), int(distance * 255))

    def set_distance(self, distance: float) -> None:
        Mix_SetDistance(self.channel, int(distance * 255))

    def set_panning(self, panning: tuple = (1.0, 1.0)) -> None:
        Mix_SetPanning(self.channel, int(panning[0] * 255), int(panning[1] * 255))

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        Mix_FreeChunk(self.chunk)
        del self.app
        self.destroyed = True
        return False

    def __del__(self) -> None:
        self.destroy()


class Music:
    def __init__(self, mixer: any, path: str) -> None:
        self.destroyed = True
        self.app = mixer.app
        self.path = path
        self.music = Mix_LoadMUS(self.app.stb(path))
        if not self.music:
            self.app.raise_error(Mix_GetError)
        self.type = {
            MUS_NONE: 'NONE',
            MUS_CMD: 'CMD',
            MUS_WAV: 'WAV',
            MUS_MOD: 'MOD',
            MUS_MID: 'MID',
            MUS_OGG: 'OGG',
            MUS_MP3: 'MP3',
            MUS_MP3_MAD_UNUSED: 'MP3_MAD_UNUSED',
            MUS_FLAC: 'FLAC',
            MUS_MODPLUG_UNUSED: 'MODPLUG_UNUSED',
            MUS_OPUS: 'OPUS'
        }.get(Mix_GetMusicType(self.music))
        try:
            self.title = self.app.bts(Mix_GetMusicTitle(self.music))
            self.title_tag = self.app.bts(Mix_GetMusicTitleTag(self.music))
            self.artist_tag = self.app.bts(Mix_GetMusicArtistTag(self.music))
            self.album_tag = self.app.bts(Mix_GetMusicAlbumTag(self.music))
            self.copyright_tag = self.app.bts(Mix_GetMusicCopyrightTag(self.music))
            self.duration = Mix_MusicDuration(self.music)
            self.loop_start = Mix_GetMusicLoopStartTime(self.music)
            self.loop_end = Mix_GetMusicLoopEndTime(self.music)
            self.loop_length = Mix_GetMusicLoopLengthTime(self.music)
        except (RuntimeError, NameError):
            self.title = ''
            self.title_tag = ''
            self.artist_tag = ''
            self.album_tag = ''
            self.copyright_tag = ''
            self.duration = 0.0
            self.loop_start = 0.0
            self.loop_end = 0.0
            self.loop_length = 0.0
        self.destroyed = False

    def play(self, loops: int = 0, fade_in: float = 0.0, pos: float = 0.0) -> None:
        if fade_in:
            if pos:
                Mix_FadeInMusicPos(self.music, loops, int(fade_in * 1000), pos)
            else:
                Mix_FadeInMusic(self.music, loops, int(fade_in * 1000))
        else:
            Mix_PlayMusic(self.music, loops)
            if pos:
                self.set_pos(pos)

    @staticmethod
    def set_pos(pos: float) -> None:
        Mix_SetMusicPosition(pos)

    def get_pos(self) -> float:
        return Mix_GetMusicPosition(self.music)

    @staticmethod
    def mod_jump_to_order(order: int) -> None:
        Mix_ModMusicJumpToOrder(order)

    @staticmethod
    def pause() -> None:
        Mix_PauseMusic()

    @staticmethod
    def resume() -> None:
        Mix_ResumeMusic()

    @staticmethod
    def rewind() -> None:
        Mix_RewindMusic()

    @staticmethod
    def is_paused() -> bool:
        return bool(Mix_PausedMusic())

    @staticmethod
    def is_playing() -> bool:
        return bool(Mix_PlayingMusic())

    @staticmethod
    def get_fading_status() -> str:
        status = Mix_FadingMusic()
        if status == MIX_FADING_OUT:
            return 'out'
        if status == MIX_FADING_IN:
            return 'in'
        return 'no'

    @staticmethod
    def stop() -> None:
        Mix_HaltMusic()

    @staticmethod
    def fade_out(fade_out: float) -> None:
        Mix_FadeOutMusic(int(fade_out * 1000))

    @staticmethod
    def set_volume(volume: float = 1.0) -> None:
        Mix_VolumeMusic(int(volume * MIX_MAX_VOLUME))

    def get_volume(self) -> float:
        return Mix_GetMusicVolume(self.music) / MIX_MAX_VOLUME

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        Mix_FreeMusic(self.music)
        del self.app
        self.destroyed = True
        return False

    def __del__(self) -> None:
        self.destroy()


class Mixer:
    def __init__(
            self,
            app: any,
            freq: float = float(MIX_DEFAULT_FREQUENCY),
            audio_format: str = 'default',
            num_channels: int = MIX_DEFAULT_CHANNELS,
            chunk_size: int = 2048,
            open_device: bool = True,
            device: str = None
    ) -> None:
        self.destroyed = True
        self.app = app
        try:
            self.format_map = {
                'u8': AUDIO_U8,
                's8': AUDIO_S8,
                'u16lsb': AUDIO_U16LSB,
                's16lsb': AUDIO_S16LSB,
                'u16msb': AUDIO_U16MSB,
                's16msb': AUDIO_S16MSB,
                'u16': AUDIO_U16,
                's16': AUDIO_S16,
                's32lsb': AUDIO_S32LSB,
                's32msb': AUDIO_S32MSB,
                's32': AUDIO_S32,
                'f32lsb': AUDIO_F32LSB,
                'f32msb': AUDIO_F32MSB,
                'f32': AUDIO_F32,
                'u16sys': AUDIO_U16SYS,
                's16sys': AUDIO_S16SYS,
                's32sys': AUDIO_S32SYS,
                'f32sys': AUDIO_F32SYS,
                'default': MIX_DEFAULT_FORMAT
            }
        except NameError:
            self.format_map = {
                'u8': 8,
                's8': 32776,
                'u16lsb': 16,
                's16lsb': 32784,
                'u16msb': 4112,
                's16msb': 36880,
                'u16': 16,
                's16': 32784,
                's32lsb': 32800,
                's32msb': 36896,
                's32': 32800,
                'f32lsb': 33056,
                'f32msb': 37152,
                'f32': 33056,
                'u16sys': 16,
                's16sys': 32784,
                's32sys': 32800,
                'f32sys': 33056,
                'default': 32784 if SDL_BYTEORDER == SDL_LIL_ENDIAN else 36880
            }
        self.freq = int(freq)
        self.format = self.format_map[audio_format.lower()]
        self.channels = num_channels
        self.chunk_size = chunk_size
        self.version = self.get_version()
        self.device_opened = open_device
        if device and open_device:
            if Mix_OpenAudioDevice(
                    self.freq, self.format, self.channels, self.chunk_size, app.stb(device), SDL_AUDIO_ALLOW_ANY_CHANGE
            ) < 0:
                app.raise_error(Mix_GetError)
        elif open_device:
            if Mix_OpenAudio(self.freq, self.format, self.channels, self.chunk_size) < 0:
                app.raise_error(Mix_GetError)
        self.channels = Mix_AllocateChannels(-1)
        self.music_decoders = [self.app.bts(Mix_GetMusicDecoder(_x)) for _x in range(Mix_GetNumMusicDecoders())]
        self.chunk_decoders = [self.app.bts(Mix_GetChunkDecoder(_x)) for _x in range(Mix_GetNumChunkDecoders())]
        self.destroyed = False

    @staticmethod
    def set_master_volume(volume: float = 1.0) -> None:
        Mix_MasterVolume(int(volume * MIX_MAX_VOLUME))

    def allocate_channels(self, count: int) -> None:
        self.channels = Mix_AllocateChannels(count)

    def set_sound_fonts(self, paths: tuple) -> None:
        Mix_SetSoundFonts(self.app.stb(';'.join(paths)))

    def set_timidity_config(self, path: str) -> None:
        Mix_SetTimidityCfg(self.app.stb(path))

    @staticmethod
    def get_version() -> tuple:
        ver = Mix_Linked_Version().contents
        return ver.major, ver.minor, ver.patch

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        if self.device_opened:
            self.device_opened = False
            Mix_CloseAudio()
        del self.app
        self.destroyed = True
        return False

    def __del__(self) -> None:
        self.destroy()



import threading
from sdl2 import *
try:
    from sdl2.sdlimage import *
except:  # noqa
    pass


class Loader:
    def __init__(self, app: any, to_load: any, extra_data: any = None) -> None:
        self.destroyed = True
        self.app = app
        self.extra_data = extra_data
        self.destroyed = False
        self.finished = False
        self.progress = 0.0
        self.counter = 0
        self.total_len = len(to_load)
        self.to_load = list(to_load)
        self.result = []

    def reset(self) -> None:
        self.finished = False
        self.progress = 0.0
        self.counter = 0
        self.total_len = len(self.to_load)
        self.result.clear()

    def run(self) -> None:
        threading.Thread(target=self.thread).start()

    def thread(self) -> None:
        while self.to_load:
            self.result.append(self.load(self.to_load.pop(0)))
            self.counter += 1
            self.progress = self.counter / self.total_len
        self.finished = True

    def load(self, to_load: any) -> any:
        pass
    
    def call_on_finish(self, cb: any) -> None:
        while not self.finished:
            pass
        cb()

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        del self.app
        del self.load
        self.destroyed = True
        return False

    def __del__(self) -> None:
        self.destroy()



import ctypes
import struct
from .surface import Surface
from sdl2 import *

try:
    from sdl2.sdlttf import *
except:  # noqa
    pass

try:
    TTF_HINTING_LIGHT_SUBPIXEL = TTF_HINTING_LIGHT_SUBPIXEL
except NameError:
    TTF_HINTING_LIGHT_SUBPIXEL = 4


class TTF:
    def __init__(self, app: any, path: str, size: float, index: int = 0) -> None:
        self.destroyed = True
        self.app = app
        self.size = int(size)
        self.scale = (1.0, 1.0)
        self.encoding = app.encoding
        self.unicode_encoding = 'utf-16'
        self.hint_map = {
            'normal': TTF_HINTING_NORMAL,
            'light': TTF_HINTING_LIGHT,
            'mono': TTF_HINTING_MONO,
            'none': TTF_HINTING_NONE,
            'light_subpixel': TTF_HINTING_LIGHT_SUBPIXEL
        }
        self.r_hint_map = {b: a for a, b in self.hint_map.items()}
        self.font = TTF_OpenFontIndex(app.stb(path), self.size, index)
        if not self.font:
            app.raise_error(TTF_GetError)
        self.normal = False
        self.bold = False
        self.italic = False
        self.underline = False
        self.strike_through = False
        self.outline = TTF_GetFontOutline(self.font)
        self.faces = TTF_FontFaces(self.font)
        self.fixed_width = bool(TTF_FontFaceIsFixedWidth(self.font))
        self.face = app.bts(TTF_FontFaceFamilyName(self.font))
        self.face_style = app.bts(TTF_FontFaceStyleName(self.font))
        try:
            self.sdf = bool(TTF_GetFontSDF(self.font))
        except NameError:
            self.sdf = False
        self.hinting = self.r_hint_map[TTF_GetFontHinting(self.font)]
        self.height = 0
        self.ascent = 0
        self.descent = 0
        self.line_skip = 0
        self.kerning = False
        self.wrapped_align = 'left'
        self.update_styles()
        self.update_vars()
        self.destroyed = False

    def set_sdf(self, sdf: bool) -> None:
        TTF_SetFontSDF(self.font, sdf)

    def set_script(self, script: str) -> None:
        if TTF_SetFontScriptName(self.font, self.app.stb(script)) < 0:
            self.app.raise_error(TTF_GetError)

    def set_direction(self, direction: str = 'ltr') -> None:
        if direction == 'rtl':
            TTF_SetFontDirection(self.font, TTF_DIRECTION_RTL)
        elif direction == 'ttb':
            TTF_SetFontDirection(self.font, TTF_DIRECTION_TTB)
        elif direction == 'btt':
            TTF_SetFontDirection(self.font, TTF_DIRECTION_BTT)
        else:
            TTF_SetFontDirection(self.font, TTF_DIRECTION_LTR)

    def char_info(self, char: str) -> tuple:
        x1_ptr, y1_ptr, x2_ptr, y2_ptr, advance_ptr = ctypes.c_int(), ctypes.c_int(), ctypes.c_int(),\
            ctypes.c_int(), ctypes.c_int()
        TTF_GlyphMetrics32(self.font, ord(char), x1_ptr, y1_ptr, x2_ptr, y2_ptr, advance_ptr)
        return x1_ptr.value, y1_ptr.value, x2_ptr.value, y2_ptr.value, advance_ptr.value

    def has_char(self, char: str) -> bool:
        return bool(TTF_GlyphIsProvided32(self.font, ord(char)))

    def update_vars(self) -> None:
        self.height = TTF_FontHeight(self.font)
        self.ascent = TTF_FontAscent(self.font)
        self.descent = TTF_FontDescent(self.font)
        self.line_skip = TTF_FontLineSkip(self.font)
        self.kerning = bool(TTF_GetFontKerning(self.font))

    def set_kerning(self, kerning: bool) -> None:
        self.kerning = kerning
        TTF_SetFontKerning(self.font, kerning)

    def set_wrapped_align(self, wrapped_align: str) -> None:
        self.wrapped_align = wrapped_align
        if wrapped_align == 'center':
            TTF_SetFontWrappedAlign(self.font, TTF_WRAPPED_ALIGN_CENTER)
        elif wrapped_align == 'right':
            TTF_SetFontWrappedAlign(self.font, TTF_WRAPPED_ALIGN_RIGHT)
        else:
            TTF_SetFontWrappedAlign(self.font, TTF_WRAPPED_ALIGN_LEFT)

    def set_hinting(self, hinting: str = 'normal') -> None:
        self.hinting = hinting
        TTF_SetFontHinting(self.font, self.hint_map[hinting])

    def set_size(self, size: float) -> None:
        self.size = int(size)
        TTF_SetFontSizeDPI(
            self.font, self.size,
            int(72 * self.scale[0]), int(72 * self.scale[1])
        )
        self.update_vars()

    def set_scale(self, scale: any = (1.0, 1.0)) -> None:
        self.scale = scale
        TTF_SetFontSizeDPI(
            self.font, self.size,
            int(72 * self.scale[0]), int(72 * self.scale[1])
        )
        self.update_vars()

    def render_unicode_wrapped(
            self, text: str, fg: any, bg: any = None, blend: bool = False, wrap_length: float = 0.0
    ) -> Surface:
        if blend:
            if bg:
                surf = TTF_RenderUNICODE_Shaded_Wrapped(
                    self.font,
                    self.encode_unicode(text),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                    SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255),
                    int(wrap_length)
                )
            else:
                surf = TTF_RenderUNICODE_Blended_Wrapped(
                    self.font,
                    self.encode_unicode(text),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                    int(wrap_length)
                )
        elif bg:
            surf = TTF_RenderUNICODE_LCD_Wrapped(
                self.font,
                self.encode_unicode(text),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255),
                int(wrap_length)
            )
        else:
            surf = TTF_RenderUNICODE_Solid_Wrapped(
                self.font,
                self.encode_unicode(text),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                int(wrap_length)
            )
        return Surface(surf, self.app)

    def render_unicode(self, text: str, fg: any, bg: any = None, blend: bool = False) -> Surface:
        if blend:
            if bg:
                surf = TTF_RenderUNICODE_Shaded(
                    self.font,
                    self.encode_unicode(text),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                    SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255)
                )
            else:
                surf = TTF_RenderUNICODE_Blended(
                    self.font,
                    self.encode_unicode(text),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255)
                )
        elif bg:
            surf = TTF_RenderUNICODE_LCD(
                self.font,
                self.encode_unicode(text),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255)
            )
        else:
            surf = TTF_RenderUNICODE_Solid(
                self.font,
                self.encode_unicode(text),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255)
            )
        return Surface(surf, self.app)

    def render_utf8_wrapped(
            self, text: str, fg: any, bg: any = None, blend: bool = False, wrap_length: float = 0.0
    ) -> Surface:
        if blend:
            if bg:
                surf = TTF_RenderUTF8_Shaded_Wrapped(
                    self.font,
                    self.app.stb(text, self.encoding),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                    SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255),
                    int(wrap_length)
                )
            else:
                surf = TTF_RenderUTF8_Blended_Wrapped(
                    self.font,
                    self.app.stb(text, self.encoding),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                    int(wrap_length)
                )
        elif bg:
            surf = TTF_RenderUTF8_LCD_Wrapped(
                self.font,
                self.app.stb(text, self.encoding),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255),
                int(wrap_length)
            )
        else:
            surf = TTF_RenderUTF8_Solid_Wrapped(
                self.font,
                self.app.stb(text, self.encoding),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                int(wrap_length)
            )
        return Surface(surf, self.app)

    def render_utf8(self, text: str, fg: any, bg: any = None, blend: bool = False) -> Surface:
        if blend:
            if bg:
                surf = TTF_RenderUTF8_Shaded(
                    self.font,
                    self.app.stb(text, self.encoding),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                    SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255)
                )
            else:
                surf = TTF_RenderUTF8_Blended(
                    self.font,
                    self.app.stb(text, self.encoding),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255)
                )
        elif bg:
            surf = TTF_RenderUTF8_LCD(
                self.font,
                self.app.stb(text, self.encoding),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255)
            )
        else:
            surf = TTF_RenderUTF8_Solid(
                self.font,
                self.app.stb(text, self.encoding),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255)
            )
        return Surface(surf, self.app)

    def render_text_wrapped(
            self, text: str, fg: any, bg: any = None, blend: bool = False, wrap_length: float = 0.0
    ) -> Surface:
        if blend:
            if bg:
                surf = TTF_RenderText_Shaded_Wrapped(
                    self.font,
                    self.app.stb(text, self.encoding),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                    SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255),
                    int(wrap_length)
                )
            else:
                surf = TTF_RenderText_Blended_Wrapped(
                    self.font,
                    self.app.stb(text, self.encoding),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                    int(wrap_length)
                )
        elif bg:
            surf = TTF_RenderText_LCD_Wrapped(
                self.font,
                self.app.stb(text, self.encoding),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255),
                int(wrap_length)
            )
        else:
            surf = TTF_RenderText_Solid_Wrapped(
                self.font,
                self.app.stb(text, self.encoding),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                int(wrap_length)
            )
        return Surface(surf, self.app)

    def render_text(self, text: str, fg: any, bg: any = None, blend: bool = False) -> Surface:
        if blend:
            if bg:
                surf = TTF_RenderText_Shaded(
                    self.font,
                    self.app.stb(text, self.encoding),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                    SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255)
                )
            else:
                surf = TTF_RenderText_Blended(
                    self.font,
                    self.app.stb(text, self.encoding),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255)
                )
        elif bg:
            surf = TTF_RenderText_LCD(
                self.font,
                self.app.stb(text, self.encoding),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255)
            )
        else:
            surf = TTF_RenderText_Solid(
                self.font,
                self.app.stb(text, self.encoding),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255)
            )
        return Surface(surf, self.app)

    def render_char(self, char: str, fg: any, bg: any = None, blend: bool = False) -> Surface:
        if blend:
            if bg:
                surf = TTF_RenderGlyph32_Shaded(
                    self.font,
                    ord(char),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                    SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255)
                )
            else:
                surf = TTF_RenderGlyph32_Blended(
                    self.font,
                    ord(char),
                    SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255)
                )
        elif bg:
            surf = TTF_RenderGlyph32_LCD(
                self.font,
                ord(char),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255),
                SDL_Color(int(bg[0]), int(bg[1]), int(bg[2]), int(bg[3]) if len(bg) > 3 else 255)
            )
        else:
            surf = TTF_RenderGlyph32_Solid(
                self.font,
                ord(char),
                SDL_Color(int(fg[0]), int(fg[1]), int(fg[2]), int(fg[3]) if len(fg) > 3 else 255)
            )
        return Surface(surf, self.app)

    def get_utf8_size(self, text: str) -> tuple:
        w_ptr, h_ptr = ctypes.c_int(), ctypes.c_int()
        TTF_SizeUTF8(self.font, self.app.stb(text, self.encoding), w_ptr, h_ptr)
        return w_ptr.value, h_ptr.value

    def get_unicode_size(self, text: str) -> tuple:
        w_ptr, h_ptr = ctypes.c_int(), ctypes.c_int()
        TTF_SizeUNICODE(self.font, self.encode_unicode(text), w_ptr, h_ptr)
        return w_ptr.value, h_ptr.value

    def encode_unicode(self, text: str) -> any:
        strlen = len(text) + 1
        intstr = struct.unpack('H' * strlen, text.encode(self.unicode_encoding))
        intstr = intstr + (0, )
        return (ctypes.c_uint16 * (strlen + 1))(*intstr)

    def set_outline(self, outline: float = 0.0) -> None:
        self.outline = int(outline)
        TTF_SetFontOutline(self.font, self.outline)

    def set_strike_through(self, enabled: bool) -> None:
        self.strike_through = enabled
        self.update_ttf_styles()
        self.update_styles()

    def set_underline(self, enabled: bool) -> None:
        self.underline = enabled
        self.update_ttf_styles()
        self.update_styles()

    def set_italic(self, enabled: bool) -> None:
        self.italic = enabled
        self.update_ttf_styles()
        self.update_styles()

    def set_bold(self, enabled: bool) -> None:
        self.bold = enabled
        self.update_ttf_styles()
        self.update_styles()

    def set_normal(self, enabled: bool) -> None:
        self.normal = enabled
        self.update_ttf_styles()
        self.update_styles()

    def update_ttf_styles(self) -> None:
        styles = 0
        if self.normal:
            styles |= TTF_STYLE_NORMAL
        if self.bold:
            styles |= TTF_STYLE_BOLD
        if self.italic:
            styles |= TTF_STYLE_ITALIC
        if self.underline:
            styles |= TTF_STYLE_UNDERLINE
        if self.strike_through:
            styles |= TTF_STYLE_STRIKETHROUGH
        TTF_SetFontStyle(self.font, styles)

    def update_styles(self) -> None:
        styles = TTF_GetFontStyle(self.font)
        self.normal = bool(styles & TTF_STYLE_NORMAL)
        self.bold = bool(styles & TTF_STYLE_BOLD)
        self.italic = bool(styles & TTF_STYLE_ITALIC)
        self.underline = bool(styles & TTF_STYLE_UNDERLINE)
        self.strike_through = bool(styles & TTF_STYLE_STRIKETHROUGH)

    @staticmethod
    def get_version() -> tuple:
        ver = TTF_Linked_Version().contents
        return ver.major, ver.minor, ver.patch

    @staticmethod
    def get_freetype_version() -> tuple:
        major_ptr, minor_ptr, patch_ptr = ctypes.c_int(), ctypes.c_int(), ctypes.c_int()
        TTF_GetFreeTypeVersion(major_ptr, minor_ptr, patch_ptr)
        return major_ptr.value, minor_ptr.value, patch_ptr.value

    @staticmethod
    def get_harfbuzz_version() -> tuple:
        major_ptr, minor_ptr, patch_ptr = ctypes.c_int(), ctypes.c_int(), ctypes.c_int()
        TTF_GetHarfBuzzVersion(major_ptr, minor_ptr, patch_ptr)
        return major_ptr.value, minor_ptr.value, patch_ptr.value

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        TTF_CloseFont(self.font)
        del self.app
        self.destroyed = True
        return False

    def __del__(self) -> None:
        self.destroy()



import ctypes
from .video import PixelFormat
from sdl2 import *


class Texture:
    def __init__(self, texture: SDL_Texture, renderer: any) -> None:
        self.destroyed = True
        if not texture:
            renderer.app.raise_error()
        self.app = renderer.app
        self.renderer = renderer
        self.texture = texture
        self.scale_mode = self.get_scale_mode()
        self.color_mod = self.get_color_mod()
        self.alpha_mod = self.get_alpha_mod()
        self.blend_mode = self.get_blend_mode_int()
        self.destroyed = False

    def get_scale_mode(self) -> str:
        scale_mode_ptr = ctypes.c_int()
        try:
            SDL_GetTextureScaleMode(self.texture, scale_mode_ptr)
        except NameError:
            return 'best'
        if scale_mode_ptr.value == SDL_ScaleModeBest:
            return 'best'
        if scale_mode_ptr.value == SDL_ScaleModeNearest:
            return 'nearest'
        if scale_mode_ptr.value == SDL_ScaleModeLinear:
            return 'linear'
        return 'unknown'

    def set_integer_scale(self, enabled: bool) -> None:
        SDL_RenderSetIntegerScale(self.renderer, enabled)

    def get_integer_scale(self) -> bool:
        return bool(SDL_RenderGetIntegerScale(self.renderer))

    def set_scale_mode(self, scale_mode: str) -> None:
        if scale_mode == 'best':
            SDL_SetTextureScaleMode(self.texture, SDL_ScaleModeBest)
        elif scale_mode == 'nearest':
            SDL_SetTextureScaleMode(self.texture, SDL_ScaleModeNearest)
        elif scale_mode == 'linear':
            SDL_SetTextureScaleMode(self.texture, SDL_ScaleModeLinear)
        self.scale_mode = self.get_scale_mode()

    def set_color_mod(self, color_mod: any) -> None:
        SDL_SetTextureColorMod(self.texture, int(color_mod[0]), int(color_mod[1]), int(color_mod[2]))
        self.color_mod = self.get_color_mod()

    def get_color_mod(self) -> tuple:
        r_ptr, g_ptr, b_ptr = ctypes.c_ubyte(), ctypes.c_ubyte(), ctypes.c_ubyte()
        SDL_GetTextureColorMod(self.texture, r_ptr, g_ptr, b_ptr)
        return r_ptr.value, g_ptr.value, b_ptr.value

    def update_blend_mode_by_alpha(self) -> None:
        SDL_SetTextureBlendMode(self.texture, SDL_BLENDMODE_NONE if self.alpha_mod >= 255 else SDL_BLENDMODE_NONE)

    def get_blend_mode(self) -> str:
        return self.app.r_blend_map[self.blend_mode]

    def set_blend_mode(self, blend_mode: str) -> None:
        self.set_blend_mode_int(self.app.blend_map[blend_mode])

    def get_blend_mode_int(self) -> int:
        try:
            blend_mode_ptr = ctypes.c_long()
            SDL_GetTextureBlendMode(self.texture, blend_mode_ptr)
        except ctypes.ArgumentError:
            blend_mode_ptr = ctypes.c_int()
            SDL_GetTextureBlendMode(self.texture, blend_mode_ptr)
        return blend_mode_ptr.value

    def set_blend_mode_int(self, blend_mode: int) -> None:
        SDL_SetTextureBlendMode(self.texture, blend_mode)
        self.blend_mode = self.get_blend_mode_int()

    def get_alpha_mod(self) -> int:
        alpha_ptr = ctypes.c_ubyte()
        SDL_GetTextureAlphaMod(self.texture, alpha_ptr)
        return alpha_ptr.value

    def set_alpha_mod(self, alpha_mod: float) -> None:
        SDL_SetTextureAlphaMod(self.texture, int(alpha_mod))
        self.alpha_mod = self.get_alpha_mod()
        self.update_blend_mode_by_alpha()

    def get_size(self) -> tuple:
        w_ptr, h_ptr = ctypes.c_int(), ctypes.c_int()
        SDL_QueryTexture(self.texture, None, None, w_ptr, h_ptr)
        return w_ptr.value, h_ptr.value

    def get_w(self) -> int:
        w_ptr = ctypes.c_int()
        SDL_QueryTexture(self.texture, None, None, w_ptr, None)
        return w_ptr.value

    def get_h(self) -> int:
        h_ptr = ctypes.c_int()
        SDL_QueryTexture(self.texture, None, None, None, h_ptr)
        return h_ptr.value

    def get_access_type(self) -> str:
        access_ptr = ctypes.c_int()
        SDL_QueryTexture(self.texture, None, access_ptr, None, None)
        if access_ptr.value == SDL_TEXTUREACCESS_STREAMING:
            return 'streaming'
        if access_ptr.value == SDL_TEXTUREACCESS_TARGET:
            return 'target'
        if access_ptr.value == SDL_TEXTUREACCESS_STATIC:
            return 'static'
        return 'none'

    def get_format(self) -> PixelFormat:
        format_ptr = ctypes.c_uint32()
        SDL_QueryTexture(self.texture, format_ptr, None, None, None)
        return PixelFormat(format_ptr.value, self.app)

    def lock(self) -> None:
        SDL_LockTexture(self.texture)

    def unlock(self) -> None:
        SDL_UnlockTexture(self.texture)

    def update(self, data: any, pitch: int, update_rect: any = None) -> None:
        SDL_UpdateTexture(
            self.texture,
            update_rect and SDL_Rect(int(update_rect[0]), int(update_rect[1]),
                                     int(update_rect[2]), int(update_rect[3])),
            data,
            pitch
        )

    def update_yuv(
            self, y_plane: any, y_pitch: int, u_plane: any, u_pitch: int,
            v_plane: any, v_pitch: int, update_rect: any = None
    ) -> None:
        SDL_UpdateYUVTexture(
            self.texture,
            update_rect and SDL_Rect(int(update_rect[0]), int(update_rect[1]),
                                     int(update_rect[2]), int(update_rect[3])),
            y_plane, y_pitch, u_plane, u_pitch, v_plane, v_pitch
        )

    def update_nv(self, y_plane: any, y_pitch: int, uv_plane: any, uv_pitch: int, update_rect: any = None) -> None:
        SDL_UpdateNVTexture(
            self.texture,
            update_rect and SDL_Rect(int(update_rect[0]), int(update_rect[1]),
                                     int(update_rect[2]), int(update_rect[3])),
            y_plane, y_pitch, uv_plane, uv_pitch
        )

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        SDL_DestroyTexture(self.texture)
        self.destroyed = True
        del self.renderer
        del self.app
        return False

    def __del__(self) -> None:
        self.destroy()



from sdl2 import *


class Clock:
    def __init__(self) -> None:
        self.freq = SDL_GetPerformanceFrequency()
        self.delta = 0.0
        self.speed_hack = 1.0
        self.last_tick = 0
        self.fps_limit = 0

    @staticmethod
    def get_time() -> float:
        return SDL_GetTicks64() / 1000

    def sleep(self, time: float) -> None:
        wait_for = self.get_time() + time
        while self.get_time() < wait_for:
            continue

    def reset(self) -> None:
        self.last_tick = SDL_GetPerformanceCounter()

    def tick(self) -> bool:
        now = SDL_GetPerformanceCounter()
        if self.fps_limit and self.freq > self.fps_limit * (now - self.last_tick):
            return False
        self.delta = (now - self.last_tick) / self.freq * self.speed_hack
        self.last_tick = now
        return True

    def get_fps(self) -> int:
        try:
            return int(self.speed_hack / self.delta)
        except ZeroDivisionError:
            return self.freq

    def get_fps_float(self) -> float:
        try:
            return self.speed_hack / self.delta
        except ZeroDivisionError:
            return float(self.freq)


class Timer:
    def __init__(
            self,
            duration: float,
            repeat: bool = True,
            enabled: bool = False,
            smooth: bool = False
    ) -> None:
        self.duration = duration
        self.repeat = repeat
        self.enabled = enabled
        self.counter = 0.0
        self.triggered = 0
        self.smooth = smooth

    def reset(self) -> None:
        self.triggered = 0
        self.counter = 0.0

    def run(self) -> None:
        self.enabled = True

    def pause(self) -> None:
        self.enabled = False

    def stop(self) -> None:
        self.pause()
        self.reset()

    def tick(self, dt: float) -> None:
        if not self.enabled:
            return
        self.counter += dt
        if self.smooth:
            if self.counter >= self.duration:
                self.counter -= self.duration
                self.triggered += 1
            return
        while self.counter >= self.duration:
            self.counter -= self.duration
            self.triggered += 1


class Animation:
    def __init__(
            self,
            duration: float = 0.0,
            repeat: bool = False,
            enabled: bool = False
    ) -> None:
        self.duration = duration
        self.repeat = repeat
        self.enabled = enabled
        self.counter = 0.0
        self.value: any = None

    def reset(self) -> None:
        self.counter = 0.0

    def run(self) -> None:
        self.enabled = True

    def stop(self) -> None:
        self.pause()
        self.reset()

    def pause(self) -> None:
        self.enabled = False

    def tick(self, dt: float) -> None:
        if not self.enabled:
            return
        self.counter += dt
        if self.duration and self.counter >= self.duration:
            if self.repeat:
                self.counter -= self.duration
            else:
                self.value = self.calc(self.duration)
                return self.stop()
        self.value = self.calc(self.counter)

    @staticmethod
    def calc(counter: float) -> any:
        return counter



from sdl2 import *


sdl_dir = dir()


class SDLVersion:
    def __init__(self, ver: SDL_version) -> None:
        self.ver = ver
        self.major = ver.major
        self.minor = ver.minor
        self.patch = ver.patch
        self.num = (self.major << 24) | (self.minor << 8) | (self.patch << 0)

    def __str__(self) -> str:
        return f'{self.major}.{self.minor}.{self.patch}'

    def __int__(self) -> int:
        return self.num
